# -*- coding: utf-8 -*-
"""
    __author__ = 'Zehra Sarica'
    __email__ = ['sarica16@itu.edu.tr','zehraacar559@gmail.com']
"""

import os.path
import re

import matplotlib.cm as cm
import matplotlib.colors as mcolors
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import plotly.express as px
import seaborn as sns
from matplotlib.ticker import MaxNLocator

from rinpy.constants import FREQUENCY_HIGH_PERCENTAGE_TEMPLATE, CHAIN_ID, RESIDUE_NUMBER, RESIDUE_NAME, PDB_ID, \
    CENTRALITY_SCORE, INSERTION, EXCEL_EXT, CSV_EXT, HTML_EXT, ATOM_NUMBER, SEGMENT_ID
from rinpy.style_config import FONT_STYLES, FONT_FAMILY
from rinpy.utils import CentralityType

_RESIDUE_ID = "residue_id"
_OCCURRENCES = "occurrences"
_PDBS = "pdbs"
_FREQUENCY = "frequency"
_PDB_IDS = 'pdb_ids'
_COUNT = 'count'

_COLONS = [PDB_ID, RESIDUE_NAME, CHAIN_ID, RESIDUE_NUMBER, INSERTION, SEGMENT_ID, CENTRALITY_SCORE]


class QuantileAnalyzer:
    def __init__(self, high_percentage_dict=None, centrality_type=CentralityType.BET, destination_output_path=None):
        self.high_percentage_dict = high_percentage_dict
        self.centrality_type = centrality_type
        self.destination_output_path = destination_output_path
        self.outfile_df = pd.DataFrame(columns=_COLONS)
        self.merge_files_and_convert_to_df()

    def merge_files_and_convert_to_df(self):
        """ Merge all files and convert them into a dataframe."""
        file_row_list = []
        for pdb_id, out_file_path in self.high_percentage_dict.items():
            file_out_df = pd.read_csv(out_file_path,
                                      sep=";",
                                      names=[ATOM_NUMBER, CENTRALITY_SCORE, RESIDUE_NAME, CHAIN_ID, RESIDUE_NUMBER,
                                             INSERTION, SEGMENT_ID])
            for _, row in file_out_df.iterrows():
                file_row_list.append({
                    PDB_ID: pdb_id,
                    RESIDUE_NAME: row[RESIDUE_NAME],
                    CHAIN_ID: row[CHAIN_ID],
                    RESIDUE_NUMBER: row[RESIDUE_NUMBER],
                    INSERTION: str(row[INSERTION]).strip() if pd.notna(row[INSERTION]) else "",
                    SEGMENT_ID: str(row[SEGMENT_ID]).strip() if pd.notna(row[SEGMENT_ID]) else "",
                    CENTRALITY_SCORE: row[CENTRALITY_SCORE]
                })

        new_data_df = pd.DataFrame(file_row_list)
        if self.outfile_df.empty:
            self.outfile_df = new_data_df
        else:
            self.outfile_df = pd.concat([self.outfile_df, new_data_df], ignore_index=True)

    def _find_frequency(self):
        """Finds the frequency of each residue along with its name, number, chain id, and insertion code"""
        return self.outfile_df.groupby([RESIDUE_NAME, RESIDUE_NUMBER, CHAIN_ID, INSERTION, SEGMENT_ID]).size()

    def run_analysis(self):
        """Runs the analysis to find frequencies and save the results to a csv file and plot bar plot and heatmaps"""
        fre_df = self._find_frequency()
        sorted_out_file_list = self._write_to_file(fre_df)
        self._save_bar_plot(sorted_out_file_list)
        self._save_heatmap(sorted_out_file_list)
        self._save_heatmap_interactive(sorted_out_file_list)

    def _write_to_file(self, fre_df: pd.DataFrame):
        """Writes the results to a csv and an Excel files"""
        fre_df = fre_df.sort_values(ascending=False)
        out_file_list = []
        for index, value in fre_df.items():
            row_df = self.outfile_df[
                (self.outfile_df[RESIDUE_NAME] == index[0]) & (self.outfile_df[RESIDUE_NUMBER] == index[1]) & (
                        self.outfile_df[CHAIN_ID] == index[2]) & (self.outfile_df[INSERTION] == index[3]) & (
                        self.outfile_df[SEGMENT_ID] == index[4])]
            pdb_ids = []
            for r in row_df.iterrows():
                pdb_ids.append(r[1][PDB_ID])

            out_file_list.append({
                RESIDUE_NAME: index[0],
                RESIDUE_NUMBER: index[1],
                CHAIN_ID: index[2],
                INSERTION: index[3],
                SEGMENT_ID: index[4],
                _COUNT: value,
                _PDB_IDS: pdb_ids
            })

        sorted_out_file_list = sorted(out_file_list,
                                      key=lambda x: (x[CHAIN_ID], x[RESIDUE_NUMBER], x[RESIDUE_NAME], x[_COUNT]))

        out_file_df = pd.DataFrame(sorted_out_file_list)
        out_file_df[_PDB_IDS] = out_file_df[_PDB_IDS].astype(str).str.replace("'", '')
        output_filename = os.path.join(self.destination_output_path, FREQUENCY_HIGH_PERCENTAGE_TEMPLATE.format(
            type=self.centrality_type.display_name))

        self._export_to_frequency_with_pdb_ids_to_csv_file(out_file_df=out_file_df, output_filename=output_filename)

        out_file_df[_PDB_IDS] = out_file_df[_PDB_IDS].astype(str).str.replace('[\]\[]', '')

        self._write_to_excel(out_file_df=out_file_df, output_filename=output_filename)

        self._export_to_frequency_without_pdb_ids_to_csv_file(fre_df=fre_df, output_filename=output_filename)

        return sorted_out_file_list

    @staticmethod
    def _write_to_excel(out_file_df: pd.DataFrame, output_filename: str) -> None:
        """Writes dataframe to the Excel file."""
        df_copy = out_file_df.copy()
        headers = [RESIDUE_NAME, RESIDUE_NUMBER, CHAIN_ID, INSERTION, SEGMENT_ID, _FREQUENCY, _PDBS]
        df_copy.columns = headers
        out_file_df.to_excel(f"{output_filename}{EXCEL_EXT}", index=False)

    @staticmethod
    def _export_to_frequency_with_pdb_ids_to_csv_file(out_file_df: pd.DataFrame, output_filename: str) -> None:
        max_residue_name_len = out_file_df[RESIDUE_NAME].astype(str).map(len).max()
        max_residue_num_len = out_file_df[RESIDUE_NUMBER].astype(str).map(len).max()
        max_chain_id_len = out_file_df[CHAIN_ID].astype(str).map(len).max()
        max_insertion_len = out_file_df[INSERTION].astype(str).map(len).max()
        max_segment_id_len = out_file_df[SEGMENT_ID].astype(str).map(len).max()
        max_count_len = out_file_df[_COUNT].astype(str).map(len).max()

        out_file_df[RESIDUE_NAME] = out_file_df[RESIDUE_NAME].astype(str).map(
            lambda x: x.ljust(max_residue_name_len))
        out_file_df[RESIDUE_NUMBER] = out_file_df[RESIDUE_NUMBER].astype(str).map(
            lambda x: x.rjust(max_residue_num_len))
        out_file_df[CHAIN_ID] = out_file_df[CHAIN_ID].astype(str).map(lambda x: x.ljust(max_chain_id_len))
        out_file_df[INSERTION] = out_file_df[INSERTION].astype(str).map(lambda x: x.ljust(max_insertion_len))
        out_file_df[SEGMENT_ID] = out_file_df[SEGMENT_ID].astype(str).map(lambda x: x.ljust(max_segment_id_len))
        out_file_df[_COUNT] = out_file_df[_COUNT].astype(str).map(lambda x: x.rjust(max_count_len))

        out_file_df = out_file_df.sort_values(by=[CHAIN_ID, RESIDUE_NUMBER, RESIDUE_NAME, _COUNT])

        out_file_df.to_csv(
            path_or_buf=f"{output_filename}{CSV_EXT}",
            sep='\t',
            header=False,
            index=False)

    @staticmethod
    def _export_to_frequency_without_pdb_ids_to_csv_file(fre_df: pd.DataFrame, output_filename: str) -> None:
        index_df = fre_df.index.to_frame(index=False).astype(str)

        max_residue_name_len = index_df.iloc[:, 0].map(len).max()
        max_residue_num_len = index_df.iloc[:, 1].map(len).max()
        max_chain_id_len = index_df.iloc[:, 2].map(len).max()
        max_insertion_len = index_df.iloc[:, 3].map(len).max()
        max_segment_id_len = index_df.iloc[:, 4].map(len).max()
        max_count_len = fre_df.astype(str).map(len).max()

        index_df.iloc[:, 0] = index_df.iloc[:, 0].str.ljust(max_residue_name_len)
        index_df.iloc[:, 1] = index_df.iloc[:, 1].str.rjust(max_residue_num_len)
        index_df.iloc[:, 2] = index_df.iloc[:, 2].str.ljust(max_chain_id_len)
        index_df.iloc[:, 3] = index_df.iloc[:, 3].str.ljust(max_insertion_len)
        index_df.iloc[:, 4] = index_df.iloc[:, 4].str.ljust(max_segment_id_len)
        counts_padded = fre_df.astype(str).str.rjust(max_count_len)

        final_df = pd.concat([index_df, counts_padded.reset_index(drop=True)], axis=1)

        final_df.to_csv(
            path_or_buf=f"{output_filename}_sorted_without_pdb_ids.csv",
            sep='\t',
            header=False,
            index=False)

    @staticmethod
    def _residue_tuple_sort_key(residue_tuple):
        """extract key items from residue tuple"""
        res_num, ins, seg_id = residue_tuple
        return res_num, ins or "", seg_id or ""

    def _save_bar_plot(self, sorted_out_file_list):
        """save frequency bar plot for the given high percentage centrality scores."""
        unique_chain_ids = sorted(set(entry[CHAIN_ID] for entry in sorted_out_file_list))
        colors = plt.get_cmap('jet')(np.linspace(0, 1, len(unique_chain_ids)))
        residue_counts = {chain_id: {} for chain_id in unique_chain_ids}
        for chain_id in unique_chain_ids:
            chain_entries = [entry for entry in sorted_out_file_list if entry[CHAIN_ID] == chain_id]
            chain_entries = sorted(chain_entries, key=lambda x: (x[RESIDUE_NUMBER], x[INSERTION], x[SEGMENT_ID]))

            for entry in chain_entries:
                residue_number = entry[RESIDUE_NUMBER]
                insertion = entry[INSERTION]
                segment_id = entry[SEGMENT_ID]
                count = entry[_COUNT]

                key = (residue_number, insertion, segment_id)

                if key not in residue_counts[chain_id]:
                    residue_counts[chain_id][key] = 0

                residue_counts[chain_id][key] += count

        unique_residue_numbers = set()
        for counts in residue_counts.values():
            unique_residue_numbers.update(counts.keys())

        unique_residue_numbers = sorted(unique_residue_numbers, key=self._residue_tuple_sort_key)

        x_positions = np.arange(len(unique_residue_numbers))

        bar_width = 0.2

        fig, ax = plt.subplots(figsize=(12, 6))

        for i, (chain_id, color) in enumerate(zip(unique_chain_ids, colors)):
            counts = [residue_counts[chain_id].get(r, 0) for r in unique_residue_numbers]
            x_positions_adjusted = x_positions + i * bar_width
            ax.bar(x_positions_adjusted, counts, width=bar_width, label=f"Chain {chain_id}", color=color)

        x_labels = [self._format_residue_label(num, ins, seg_id) for num, ins, seg_id in
                    sorted(unique_residue_numbers, key=self._residue_tuple_sort_key)]

        ax.set_xticks(x_positions)
        ax.set_xticklabels(x_labels, rotation=90, **{
            "fontsize": FONT_STYLES["xtick_small"]["labelsize"],
            "fontfamily": FONT_STYLES["xtick"]["fontfamily"]
        })

        has_any_seg_id = any(
            seg_id not in (None, '', "''")
            for _, _, seg_id in unique_residue_numbers
        )
        if has_any_seg_id:
            ax.set_xlabel("Residue Number (Segment ID)", fontdict=FONT_STYLES["xlabel"])
        else:
            ax.set_xlabel("Residue Number", fontdict=FONT_STYLES["xlabel"])
        ax.set_ylabel("Frequency of High Percentage Residues", fontdict=FONT_STYLES["xlabel"])
        ax.legend(loc="upper right")
        ax.legend(prop={"family": FONT_FAMILY, "size": FONT_STYLES["legend"]["fontsize"]})

        ax.yaxis.set_major_locator(MaxNLocator(integer=True, nbins="auto", min_n_ticks=2))

        plt.tight_layout()

        plot_output_filename = os.path.join(self.destination_output_path,
                                            f"frequency_high_percentage_{self.centrality_type.display_name}_bar_plot")
        plt.savefig(plot_output_filename, dpi=300)

    def _prepare_heatmap_dataframe(self, sorted_out_file_list, is_interactive: bool = False):
        """ prepare the heatmap data for the given high percentage centrality score residues."""
        residue_number_dtype = str if is_interactive else int

        total_pdb_ids = len(set(pdb_id for entry in sorted_out_file_list for pdb_id in entry[_PDB_IDS]))
        data = [[entry[RESIDUE_NAME], entry[RESIDUE_NUMBER], entry[CHAIN_ID], entry[INSERTION], entry[SEGMENT_ID],
                 entry[_COUNT], entry[_PDB_IDS]] for entry in sorted_out_file_list]
        df = pd.DataFrame(data, columns=[RESIDUE_NAME, RESIDUE_NUMBER, CHAIN_ID, INSERTION, SEGMENT_ID, _FREQUENCY,
                                         _OCCURRENCES])
        df[RESIDUE_NUMBER] = df[RESIDUE_NUMBER].astype(residue_number_dtype)
        df[_FREQUENCY] = df[_FREQUENCY].astype(int)
        df[_RESIDUE_ID] = df.apply(self._make_residue_id, axis=1)
        heatmap_data = df.pivot_table(index=CHAIN_ID, columns=_RESIDUE_ID, values=_FREQUENCY, aggfunc="sum",
                                      fill_value=0)
        return df, heatmap_data, total_pdb_ids

    def _residue_sort_key(self, residue_id: str):
        """ extracts and returns key ids from the given the residue id."""
        res_num, ins, seg_id = self._parse_residue_string(residue_id)
        return res_num, ins or "", seg_id or ""

    def _save_heatmap(self, sorted_out_file_list):
        """ saves the heatmap data as a PNG file."""
        df, heatmap_data, total_pdb_ids = self._prepare_heatmap_dataframe(
            sorted_out_file_list=sorted_out_file_list,
            is_interactive=False
        )
        heatmap_data = heatmap_data.sort_index(ascending=False)
        heatmap_data = heatmap_data[sorted(heatmap_data.columns, key=self._residue_sort_key)]

        heatmap_data = heatmap_data.apply(pd.to_numeric, errors='coerce')

        max_freq_data = int(np.nanmax(heatmap_data.values))

        max_freq = max(max_freq_data, total_pdb_ids)

        if max_freq <= 10:
            tick_vals = np.arange(0, max_freq + 1)
        else:
            step = max(1, round(max_freq / 10))
            tick_vals = list(range(0, max_freq + 1, step))
            if max_freq not in tick_vals:
                tick_vals.append(max_freq)

        custom_cmap = sns.diverging_palette(240, 10, as_cmap=True)

        plt.figure(figsize=(14, 8))
        sns_heatmap = sns.heatmap(
            heatmap_data,
            cmap=custom_cmap,
            annot=True,
            fmt="d",
            annot_kws={"size": 8},
            linewidths=0.5,
            linecolor="white",
            cbar_kws={
                "ticks": tick_vals,
                "label": "Frequency"
            },
            vmin=0,
            vmax=max_freq
        )
        for text in sns_heatmap.texts:
            text.set_rotation(90)

        sns_heatmap.set_xticks(np.arange(len(heatmap_data.columns)) + 0.5)

        formatted_columns = self._format_columns(heatmap_data=heatmap_data)
        sns_heatmap.set_xticklabels(
            formatted_columns,
            rotation=90,
            fontsize=FONT_STYLES['xtick_small_small']['labelsize'],
            fontname=FONT_FAMILY
        )

        cbar = sns_heatmap.collections[0].colorbar
        self.apply_colorbar_style(cbar, cbar_label="Frequency")

        plt.title(f"Residue Frequency Heatmap ({self.centrality_type.display_name_capitalized()})",
                  fontdict={"fontsize": 24, "fontfamily": FONT_FAMILY}, pad=12)
        font_dict = {"fontsize": 18, "fontfamily": FONT_FAMILY}
        if self._has_any_seg_id(df=df):
            plt.xlabel("Residue Number (Segment ID)", fontdict=font_dict)
        else:
            plt.xlabel("Residue Number", fontdict=font_dict)
        plt.ylabel("Chain ID", fontdict=font_dict)
        plt.xticks(rotation=90, fontsize=FONT_STYLES['xtick_medium']['labelsize'], fontname=FONT_FAMILY)
        plt.yticks(fontsize=font_dict['fontsize'], fontname=FONT_FAMILY)
        plt.tight_layout()
        plot_output_filename = os.path.join(self.destination_output_path,
                                            f"frequency_high_percentage_{self.centrality_type.display_name}_heatmap")
        plt.savefig(plot_output_filename, dpi=300, bbox_inches="tight")

    @staticmethod
    def _has_any_seg_id(df):
        segment_id = df[SEGMENT_ID]
        has_valid_seg_id = (segment_id.notna() & (segment_id != '') & (segment_id != "''"))
        return has_valid_seg_id.any()

    @staticmethod
    def _parse_residue_string(s: str):
        """ parsing the residue string into residue number, insertion, and segment_id."""
        s = s.strip()
        if ':' in s:
            segment_id, s = s.split(':', 1)
            segment_id = segment_id.strip()
        else:
            segment_id = ''
        if s.endswith("''"):
            residue_number = int(s[:-2])
            insertion = ''
        else:
            match = re.fullmatch(r"(\d+)([A-Za-z]?)", s)
            if not match:
                raise ValueError(f"Invalid residue string format: {s}")
            residue_number = int(match.group(1))
            insertion = match.group(2)

        return residue_number, insertion, segment_id

    @staticmethod
    def _make_residue_id(row):
        """ generates the residue identifier based on the residue number, insertion, and segment_id."""
        seg = row[SEGMENT_ID]
        ins = row[INSERTION]
        if pd.notna(seg) and seg != '' and pd.notna(ins) and ins != '':
            return f"{seg}:{row[RESIDUE_NUMBER]}{ins}"
        elif pd.notna(seg) and seg != '':
            return f"{seg}:{row[RESIDUE_NUMBER]}"
        elif pd.notna(ins) and ins != '':
            return f"{row[RESIDUE_NUMBER]}{ins}"
        else:
            return str(row[RESIDUE_NUMBER])

    @staticmethod
    def _format_residue_label(res_num, ins, seg_id):
        """ formats the residue representation in the x-axis of the plots."""
        has_insertion = ins not in (None, "", "''")
        base = f"{res_num}{ins}" if has_insertion else str(res_num)
        if seg_id not in (None, "", "''"):
            return f"{base} ({seg_id})"
        return base

    def _format_columns(self, heatmap_data):
        """ formats the heatmap data columns for the x-axis of the plots."""
        return [
            self._format_residue_label(*self._parse_residue_string(col))
            for col in heatmap_data.columns
        ]

    @staticmethod
    def _auto_tick_font_size(n_ticks, min_size=10, base_max=14):
        """ generates automatic font size for the tick based on the number of ticks."""
        if n_ticks <= 10:
            max_size = base_max + 2
        else:
            max_size = base_max
        scale = max(1.0, (n_ticks / 20) ** 0.5)
        size = max_size / scale
        return max(min_size, int(round(size)))

    def _save_heatmap_interactive(self, sorted_out_file_list):
        """ saves the heatmap data in the interactive HTML."""
        df, heatmap_data, total_pdb_ids = self._prepare_heatmap_dataframe(
            sorted_out_file_list=sorted_out_file_list,
            is_interactive=True
        )

        heatmap_data = heatmap_data.sort_index(ascending=False)
        heatmap_data = heatmap_data[sorted(heatmap_data.columns, key=self._residue_sort_key)]

        agg_df = df.groupby([CHAIN_ID, RESIDUE_NUMBER, INSERTION, SEGMENT_ID]).agg({
            RESIDUE_NAME: lambda x: ', '.join(sorted(set(x))),
            _OCCURRENCES: lambda x: ', '.join(sorted(set(
                pdb.strip()
                for sublist in x
                for pdb in str(sublist).strip("[]").replace('"', '').split(',')
            )))
        }).reset_index()

        agg_df[_RESIDUE_ID] = agg_df.apply(self._make_residue_id, axis=1)

        residue_matrix = agg_df.pivot(index=CHAIN_ID, columns=_RESIDUE_ID, values=RESIDUE_NAME).reindex(
            index=heatmap_data.index, columns=heatmap_data.columns, fill_value="")

        pdb_matrix = agg_df.pivot(index=CHAIN_ID, columns=_RESIDUE_ID, values=_OCCURRENCES).reindex(
            index=heatmap_data.index, columns=heatmap_data.columns, fill_value="")

        def wrap_pdb_ids(pdb_str, wrap_after=8):
            if not isinstance(pdb_str, str):
                return "None"
            ids = [p.strip() for p in pdb_str.strip("[]").replace('"', '').split(",") if p.strip()]
            if not ids:
                return "None"
            chunks = [', '.join(ids[i:i + wrap_after]) for i in range(0, len(ids), wrap_after)]
            return "<br>".join(chunks)

        custom_hovertext = heatmap_data.copy().astype(str)
        for row in heatmap_data.index:
            for col in heatmap_data.columns:
                resname = residue_matrix.loc[row, col]
                pdbs_raw = pdb_matrix.loc[row, col]
                freq = heatmap_data.loc[row, col]

                resname = resname if pd.notna(resname) else "None"
                pdbs_wrapped = wrap_pdb_ids(pdbs_raw) if pd.notna(pdbs_raw) else "None"
                residue_number, insertion, segment_id = self._parse_residue_string(col)
                custom_hovertext.loc[row, col] = (
                        f"<b>Residue Number</b>: "
                        f"{residue_number}{insertion if insertion and insertion != '' else ''}<br>"
                        + (f"<b>Segment ID</b>: {segment_id}<br>" if segment_id not in (None, '', "''") else "")
                        + f"<b>Chain ID</b>: {row}<br>"
                          f"<b>Residue Name(s)</b>: {resname}<br>"
                          f"<b>Frequency</b>: {freq}<br>"
                          f"<b>PDB IDs</b>: {pdbs_wrapped}"
                )

        custom_cmap = sns.diverging_palette(240, 10, as_cmap=True)
        custom_plotly_colorscale = self.matplotlib_to_plotly(custom_cmap)
        formatted_columns = self._format_columns(heatmap_data=heatmap_data)
        fig = px.imshow(
            heatmap_data,
            x=formatted_columns,
            color_continuous_scale=custom_plotly_colorscale,
            aspect='auto',
            labels=dict(color='Frequency'),
            title=f"Residue Frequency Heatmap ({self.centrality_type.display_name_capitalized()})",
            text_auto=True,
        )

        fig.update_layout(
            xaxis_title=("Residue Number (Segment ID)"
                         if self._has_any_seg_id(df=df)
                         else "Residue Number"),
            yaxis_title="Chain ID",
            xaxis=dict(
                tickangle=45,
                tickmode='linear',
                dtick=1,
                tickfont=dict(size=self._auto_tick_font_size(len(heatmap_data.columns))),
            ),
            font=dict(family=FONT_FAMILY, size=14),
        )
        # x-axes is broken, this is fixing that
        fig.update_xaxes(
            type="category",
            tickmode="array",
            tickvals=formatted_columns,
            ticktext=formatted_columns
        )

        max_freq_data = int(df["frequency"].max())

        max_freq = max(max_freq_data, total_pdb_ids)

        if max_freq <= 10:
            tick_vals = list(range(0, max_freq + 1))
        else:
            step = max(1, round(max_freq / 10))
            tick_vals = list(range(0, max_freq + 1, step))
            if max_freq not in tick_vals:
                tick_vals.append(max_freq)

        fig.update_coloraxes(
            colorbar=dict(
                tickmode='array',
                tickvals=tick_vals,
                ticktext=[str(i) for i in tick_vals],
                title=dict(
                    text='Frequency',
                    side='right',
                    font=dict(size=FONT_STYLES['legend']['title_fontsize'],
                              family=FONT_STYLES['legend']['fontfamily']),
                ),
            ),
            cmax=max_freq,
            cmin=0
        )

        fig.update_traces(hovertemplate="%{customdata}<extra></extra>", customdata=custom_hovertext.values)

        plot_output_filename = os.path.join(
            self.destination_output_path,
            f"frequency_high_percentage_{self.centrality_type.display_name}_heatmap{HTML_EXT}"
        )
        config = {
            "scrollZoom": True,
            "doubleClick": "reset",
            "displayModeBar": True,
            "modeBarButtonsToAdd": ["pan2d", "zoomIn2d", "zoomOut2d"]
        }
        fig.update_layout(dragmode='pan')

        fig.write_html(plot_output_filename, config=config)

    @staticmethod
    def matplotlib_to_plotly(cmap, n_colors=256):
        norm = mcolors.Normalize(vmin=0, vmax=1)
        scalar_map = cm.ScalarMappable(norm=norm, cmap=cmap)
        return [
            [i / (n_colors - 1), mcolors.to_hex(scalar_map.to_rgba(i / (n_colors - 1)))]
            for i in range(n_colors)
        ]

    @staticmethod
    def apply_colorbar_style(cbar, cbar_label=None):
        """Apply consistent style to colorbar ticks and label."""
        styles = FONT_STYLES["colorbar"]
        cbar.ax.tick_params(labelsize=styles["tick_params"]["labelsize"])
        for tick in cbar.ax.get_yticklabels():
            tick.set_fontfamily(styles["tick_params"].get("fontfamily", "sans-serif"))
        if cbar_label is not None:
            cbar.set_label(
                cbar_label,
                fontsize=18,
                fontfamily=FONT_FAMILY)


def main():
    pass


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
"""
    __author__ = 'Zehra Sarica'
    __email__ = ['sarica16@itu.edu.tr','zehraacar559@gmail.com']
"""

import logging
import os
import shutil
from enum import Enum

import pandas as pd
from biopandas.pdb import PandasPdb

from rinpy.constants import *


class PdbRecord(Enum):
    ATOM = 'ATOM',
    HETATM = 'HETATM',
    ANISOU = 'ANISOU',
    OTHERS = 'OTHERS'


class CentralityType(Enum):
    BET = 'BET',
    CLOS = 'CLOS',
    DEG = 'DEG'

    @property
    def display_name(self):
        names = {
            CentralityType.BET: "betweenness",
            CentralityType.CLOS: "closeness",
            CentralityType.DEG: "degree"
        }
        return names[self]

    def display_name_capitalized(self):
        return self.display_name.capitalize()


def create_empty_ppdb_atom_df(atom_df):
    ppdb = PandasPdb()
    ppdb.df[PdbRecord.ATOM.name] = atom_df
    return ppdb


def create_ppdb_atom_df(atom_df):
    df = pd.DataFrame(columns=PDB_COLUMNS)
    for idx, row in atom_df.iterrows():
        if 'record_name' in atom_df.columns:
            df.at[idx, 'record_name'] = row['record_name']
        else:
            df.at[idx, 'record_name'] = ''
        if 'atom_number' in atom_df.columns:
            df.at[idx, 'atom_number'] = row['atom_number']
        else:
            df.at[idx, 'atom_number'] = 0
        if 'blank_1' in atom_df.columns:
            df.at[idx, 'blank_1'] = row['blank_1']
        else:
            df.at[idx, 'blank_1'] = ''
        if 'atom_name' in atom_df.columns:
            df.at[idx, 'atom_name'] = row['atom_name']
        else:
            df.at[idx, 'atom_name'] = ''
        if 'alt_loc' in atom_df.columns:
            df.at[idx, 'alt_loc'] = row['alt_loc']
        else:
            df.at[idx, 'alt_loc'] = ''
        if 'residue_name' in atom_df.columns:
            df.at[idx, 'residue_name'] = row['residue_name']
        else:
            df.at[idx, 'residue_name'] = ''
        if 'blank_2' in atom_df.columns:
            df.at[idx, 'blank_2'] = row['blank_2']
        else:
            df.at[idx, 'blank_2'] = ''
        if 'chain_id' in atom_df.columns:
            df.at[idx, 'chain_id'] = row['chain_id']
        else:
            df.at[idx, 'chain_id'] = ''
        if 'residue_number' in atom_df.columns:
            df.at[idx, 'residue_number'] = row['residue_number']
        else:
            df.at[idx, 'residue_number'] = 0
        if 'insertion' in atom_df.columns:
            df.at[idx, 'insertion'] = row['insertion']
        else:
            df.at[idx, 'insertion'] = ''
        if 'blank_3' in atom_df.columns:
            df.at[idx, 'blank_3'] = row['blank_3']
        else:
            df.at[idx, 'blank_3'] = ''
        if 'x_coord' in atom_df.columns:
            df.at[idx, 'x_coord'] = row['x_coord']
        if 'y_coord' in atom_df.columns:
            df.at[idx, 'y_coord'] = row['y_coord']
        if 'z_coord' in atom_df.columns:
            df.at[idx, 'z_coord'] = row['z_coord']
        if 'occupancy' in atom_df.columns:
            df.at[idx, 'occupancy'] = row['occupancy']
        else:
            df.at[idx, 'occupancy'] = 0
        if 'b_factor' in atom_df.columns:
            df.at[idx, 'b_factor'] = row['b_factor']
        if 'blank_4' in atom_df.columns:
            df.at[idx, 'blank_4'] = row['blank_4']
        else:
            df.at[idx, 'blank_4'] = ''
        if 'segment_id' in atom_df.columns:
            df.at[idx, 'segment_id'] = row['segment_id']
        else:
            df.at[idx, 'segment_id'] = ''
        if 'element_symbol' in atom_df.columns:
            df.at[idx, 'element_symbol'] = row['element_symbol']
        else:
            df.at[idx, 'element_symbol'] = ''
        if 'charge' in atom_df.columns:
            df.at[idx, 'charge'] = row['charge']
    return df


def create_folder_not_exists(file_path=None):
    if not os.path.exists(file_path):
        os.makedirs(file_path)
        logging.info("The '{}' directory is created!".format(file_path))


def convert_pdb_to_pandas_pdb(pdb_path=None):
    ppdb = PandasPdb()
    ppdb.read_pdb(pdb_path)
    return ppdb


def clear_existing_file(file_path=None):
    if os.path.exists(file_path):
        with open(file_path, "r+") as file_object:
            file_object.truncate()
        file_object.close()


def write_ppdb_to_pdb_file(ppdb, out_filename):
    ppdb.to_pdb(path=out_filename,
                records=[PdbRecord.ATOM.name],
                gz=False,
                append_newline=True)


def write_ppdb_to_pdb_file_atom_and_hetatom(ppdb, out_filename):
    ppdb.to_pdb(path=out_filename,
                records=[PdbRecord.ATOM.name, PdbRecord.HETATM.name],
                gz=False,
                append_newline=True)


def write_ppdb_to_pdb_file_all(ppdb, out_filename):
    ppdb.to_pdb(path=out_filename,
                records=None,
                gz=False,
                append_newline=True)


def convert_pdb_df_to_atom_ppdb(pdb_df):
    ppdb = PandasPdb()
    ppdb.df[PdbRecord.ATOM.name] = pdb_df
    return ppdb


def write_centrality_scores_to_file(filename_path=None, score_dict=None, actual_residue_number_map=None,
                                    float_formatter='%9.7f'):
    if score_dict is None:
        score_dict = dict()
    score_dict = dict(sorted(score_dict.items()))
    with open(filename_path, 'w') as f:
        for key, value in score_dict.items():
            if actual_residue_number_map is not None:
                residue_name, chain_id, residue_number, insertion, segment_id, record_name = actual_residue_number_map[
                    key]
                if insertion:
                    insertion_out = insertion
                else:
                    insertion_out = "''"

                if segment_id:
                    segment_id_out = segment_id
                else:
                    segment_id_out = "''"

                line_formatter = "%d " + float_formatter + " %s %s %d %s %s\n"
                f.write(line_formatter % (
                    key, value, residue_name, chain_id, residue_number, insertion_out, segment_id_out))
            else:
                line_formatter = "%d " + float_formatter + "\n"
                f.write(line_formatter % (key, value))


def delete_folders_in_path(path):
    """ Deletes only folders and their contents for a given path.
    If the path is a file, no deletion occurs.

    Parameters:
        path (str): The path to check for folders to delete.
    """
    if not os.path.exists(path):
        logging.error(f"The given path does not exist: {path}")
        return

    if os.path.isdir(path):
        for item in os.listdir(path):
            item_path = os.path.join(path, item)
            if os.path.isdir(item_path):
                shutil.rmtree(item_path)


def get_df(output_file_path=None, columns: list = None, dtypes: dict = None,
           sort_keys: list = None, sep: str = r"\s+") -> pd.DataFrame:
    """ Convert the given text file into a pandas dataframe and return it.

    Parameters:
    -----------
    output_file_path : str, optional
        The file path of the output file to be converted into dataframe.

    columns : list of str, optional
        A list specifying the column names to be used in the resulting DataFrame.

    dtypes : dict, optional
        A dictionary specifying data types for the columns.
    sort_keys : list of str, optional
        A list of column names to sort the DataFrame by.
    Returns:
    --------
    pd.DataFrame
        The resulting DataFrame loaded from the given file.
    """
    if len(columns) != len(dtypes):
        raise ValueError("The number of columns must match the number of dtype specifications.")

    missing_columns = [col for col in columns if col not in dtypes]

    if missing_columns:
        raise ValueError(
            f"The following columns are not present in the dtype dictionary: {', '.join(missing_columns)}")

    df = pd.read_csv(output_file_path,
                     sep=sep,
                     engine="python",
                     names=columns)

    for col, dtype in dtypes.items():
        if col in df.columns:
            df[col] = df[col].astype(dtype)

    if sort_keys is not None:
        df = df.sort_values(by=sort_keys)

    return df


def get_atom_pdb_df(ppdb=None):
    """ return the dataframe for ATOM entries in the PPDB file

    Parameters:
    -----------
    ppdb : Biopandas file
        if None, raise an error
    """
    if ppdb is None:
        raise ValueError("ppdb must not be None. Please provide a valid biopandas file.")

    df = ppdb.df[PdbRecord.ATOM.name]
    df[ATOM_NUMBER] = df[ATOM_NUMBER].astype(int)
    df = df.sort_values(by=[ATOM_NUMBER], ascending=True)
    return df


def get_ppdb(pdb_file_path):
    """Converts the given PDB file to the biopandas PandasPdb object"""
    return convert_pdb_to_pandas_pdb(pdb_file_path)


def get_residue_id(node):
    res_num = str(node[RESIDUE_NUMBER])
    ins_code = node.get(INSERTION, '')
    return f"{res_num}{ins_code}" if ins_code not in ["", "''", None] else str(res_num)


def get_residue_id_by_tuple(residue_tuple: tuple):
    """ residue_tuple: tuple like ('HIS','A', 270, '', 'A') or ('HIS','A', 270, 'A', 'G')
    returns: str like '270', '270A', '270B,G' etc.
    """
    res_num = residue_tuple[2]
    ins_code = residue_tuple[3]
    seg_id = residue_tuple[4]

    base = f"{res_num}{ins_code}" if ins_code not in ["", "''", None] else str(res_num)

    if seg_id not in ["", "''", None]:
        return f"{base}, {seg_id}"

    return base


def get_base_pdb_df(residue_to, destination_output_path, pdb_name):
    base_pdb_df = convert_pdb_to_pandas_pdb(os.path.join(destination_output_path, pdb_name, f'{pdb_name}{PDB_EXT}'))
    atom_df = base_pdb_df.df[PdbRecord.ATOM.name]
    hetatm_df = base_pdb_df.df[PdbRecord.HETATM.name]

    atom_keys = list(
        zip(atom_df[RESIDUE_NAME], atom_df[CHAIN_ID], atom_df[RESIDUE_NUMBER], atom_df[INSERTION], atom_df[SEGMENT_ID]))
    hetatm_keys = list(
        zip(hetatm_df[RESIDUE_NAME], hetatm_df[CHAIN_ID], hetatm_df[RESIDUE_NUMBER], hetatm_df[INSERTION],
            hetatm_df[SEGMENT_ID]))

    base_pdb_df.df[PdbRecord.ATOM.name][B_FACTOR] = [residue_to.get(key, 0.0) for key in atom_keys]
    base_pdb_df.df[PdbRecord.HETATM.name][B_FACTOR] = [residue_to.get(key, 0.0) for key in hetatm_keys]

    return base_pdb_df



# -*- coding: utf-8 -*-
"""
    __author__ = 'Zehra Sarica'
    __email__ = ['sarica16@itu.edu.tr','zehraacar559@gmail.com']
"""

import sys

from rinpy.network_comparator import main as compare_main
from rinpy.rin_process import main as process_main


def main():
    # `rinpy compare ...` compare CLI
    if len(sys.argv) > 1 and sys.argv[1] == "compare":
        compare_main(sys.argv[2:])
        return
    # Default: RINProcess CLI
    process_main()


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
"""
    __author__ = 'Zehra Sarica'
    __email__ = ['sarica16@itu.edu.tr','zehraacar559@gmail.com']
"""

import argparse
import glob
import json
import logging
from datetime import datetime
from pathlib import Path

from rinpy import log_util, logging_config
from rinpy.centrality_analyzer import CentralityAnalyzer
from rinpy.centrality_pdb_mapper import CentralityPdbMapper
from rinpy.font_utils import load_fonts_once
from rinpy.hinge_analyzer import HingeAnalyzer
from rinpy.log_util import *
from rinpy.quantile_analyzer import QuantileAnalyzer
from rinpy.residue_network_builder import ResidueNetworkBuilder
from rinpy.trajectory import Trajectory
from rinpy.utils import *

_PDB_ID_LENGTH = 4
_CLUSTER_NUMBER = 'cluster_number'


class RINProcess:
    """
        RINProcess is entry module of the RinPy package to manage the entire process step by step.

        Attributes:
        -----------
        _fonts_initialized : bool
            Class-level flag to indicate whether fonts for plotting have been initialized.

        Parameters:
        -----------
        output_path : str
            Path where the results, plots, or processed files will be saved.

        input_path : str
            Path to the input PDB files.

        pdb_ids : list of str
            List of PDB IDs to be processed.

        trajectory_file : str,
            Path to a single PDB which contains multiple PDB from a molecular dynamics (MD) trajectory file.

        stride : int, default=1
            The step size for processing frames from a PDB file which contains MD snapshots.
            For example, stride=10 means every 10th frame will be processed.

        calculation_options : dict
            A dictionary containing calculation parameters such as distance cutoffs, or other options for RIN generation.

        ligand_dict : dict
            Dictionary containing hetero atoms to be included in the generated network.

        num_workers : int,
            Number of parallel workers to use for multiprocessing computations. Default is detected from the system.
    """

    _fonts_initialized = False

    def __init__(self, output_path: str = None, input_path: str = None, pdb_ids: str = None, trajectory_file=None,
                 stride=1, calculation_options=None, ligand_dict=None, num_workers: int = None):

        if not RINProcess._fonts_initialized:
            load_fonts_once()
            RINProcess._fonts_initialized = True

        self._validate_input_options(output_path=output_path,
                                     input_path=input_path,
                                     pdb_ids=pdb_ids,
                                     trajectory_file=trajectory_file)

        self.output_path = output_path
        self.input_path = input_path
        self.pdb_ids = pdb_ids
        self.trajectory_file = trajectory_file
        self.stride = stride
        self.ligand_dict = ligand_dict
        self.num_workers = num_workers

        self.calculation_options = calculation_options

        if self.calculation_options is None or not isinstance(calculation_options, dict):
            self.calculation_options = self._get_default_calculation_options()

    @staticmethod
    def _validate_input_options(output_path, input_path, pdb_ids, trajectory_file):
        """Validates the input options provided. The input can be one of the following:
            input_path, pdb_ids, trajectory_file
        """
        if output_path is None:
            raise ValueError('You must provide an output path to proceed.')

        inputs_provided = sum(x is not None for x in [input_path, pdb_ids, trajectory_file])
        if inputs_provided == 0:
            raise ValueError("You must provide exactly one input: 'input_path', 'pdb_ids', or 'trajectory_file'.")
        elif inputs_provided > 1:
            raise ValueError(
                "Only one input can be provided. Please provide only one of 'input_path', 'pdb_ids', or 'trajectory_file'.")

    @staticmethod
    def _get_default_calculation_options():
        """Returns the default calculation options from the calculation_options.json file
        in the current working directory."""
        json_file_path = os.path.join(os.path.dirname(os.path.abspath(__file__)), CALCULATION_OPTION_DEFAULT_JSON_PATH)
        try:
            with open(json_file_path, 'r') as json_file:
                data = json.load(json_file)
            if data is None:
                raise ValueError("Data in the JSON file is None. Please provide valid data.")
            return data
        except FileNotFoundError:
            print(f"File not found: {json_file_path}")
        except json.JSONDecodeError:
            print(f"Error decoding JSON in {json_file_path}")
        except ValueError as e:
            print(e)
        except Exception as e:
            print(f"An unexpected error occurred: {e}")

    @staticmethod
    @log_time("Fetching PDB IDs from Protein Data Bank (RCSB)")
    def _fetch_pdb_ids(pdb_ids, output_path=None):
        """Fetches PDB IDs from the Protein Data Bank (RCSB)."""
        total_start_time = datetime.strptime(datetime.now().strftime("%H:%M:%S"), "%H:%M:%S")
        fetched_pdb_files = []
        output_path_pdb_ids = []
        for pdb_id in pdb_ids:
            if len(pdb_id) == _PDB_ID_LENGTH:
                destination_path = os.path.join(output_path, pdb_id)
                create_folder_not_exists(destination_path)
                pdb_id_file = os.path.join(str(destination_path), pdb_id + PDB_EXT)
                if not os.path.exists(pdb_id_file):
                    ppdb = PandasPdb().fetch_pdb(pdb_id)
                    write_ppdb_to_pdb_file_all(ppdb, pdb_id_file)
                    fetched_pdb_files.append(pdb_id)
                    output_path_pdb_ids.append(pdb_id_file)
                else:
                    logging.warning(f"{pdb_id} has already fetched to {pdb_id_file}.")
                    output_path_pdb_ids.append(pdb_id_file)
            else:
                logging.error(f"{pdb_id} is NOT VALID PDB ID. It must be 4 character like 4OBE!")
        if fetched_pdb_files:
            logging.info(
                f"PDB IDS: {fetched_pdb_files} have been fetched from the protein data bank and saved to the path: {output_path}")
            total_end_time = datetime.strptime(datetime.now().strftime("%H:%M:%S"), "%H:%M:%S")
            delta = total_end_time - total_start_time
            logging.info(f"Total fetch duration is {delta.total_seconds()} seconds.")

        return output_path_pdb_ids

    @log_with_stars("Residue Network Builder")
    @log_time("Residue Network Builder")
    def _process_residue_network_builder(self, pdb_name=None, pdb_path=None, use_preprocess=True,
                                         destination_output_path=None,
                                         calculation_options: dict = None) -> tuple[ResidueNetworkBuilder, bool]:
        """Process Residue Network Builder to prepare necessary input files for Residue Interaction Network."""
        het_atom_list = None

        if self.ligand_dict is not None and pdb_name is not None and pdb_name in self.ligand_dict:
            het_atom_list = self.ligand_dict[pdb_name]

        cutoff_value = DEFAULT_CUTOFF

        if CUTOFF in calculation_options:
            option = calculation_options[CUTOFF]
            if option.get(IS_CHECKED):
                cutoff_value = option[VALUE]

        rgb = ResidueNetworkBuilder(pdb_name=pdb_name,
                                    pdb_path=pdb_path,
                                    use_preprocess=use_preprocess,
                                    cutoff=cutoff_value,
                                    destination_output_path=destination_output_path,
                                    het_atom_list=het_atom_list,
                                    remove_hydrogen=calculation_options[REMOVE_HYDROGEN][IS_CHECKED],
                                    num_workers=self.num_workers)

        is_success = rgb.calculate_contact()

        return rgb, is_success

    @log_with_stars("Centrality Analyzer")
    @log_time("Centrality Analyzer")
    def _process_centrality_analyzer(self, pdb_name: str = None, number_of_amino_acid=None,
                                     destination_output_path=None,
                                     calculation_options=None,
                                     actual_residue_number_map=None) -> CentralityAnalyzer:
        """Process centrality analysis from the generated network based on atom-atom distance of a given PDB."""

        ca = CentralityAnalyzer(pdb_name=pdb_name,
                                number_of_amino_acid=number_of_amino_acid,
                                destination_output_path=destination_output_path,
                                calculation_options=calculation_options,
                                actual_residue_number_map=actual_residue_number_map)
        ca.build_network()
        ca.calculate_all_scores(use_parallel=True)
        return ca

    @log_with_stars("Hinge Residues Analyzer")
    @log_time("Hinge Residues Analyzer")
    def _process_hinge_analyzer(self, graph, pdb_name, destination_output_path,
                                actual_residue_number_map, calculation_options) -> HingeAnalyzer:
        """Process hinges from the generate network."""
        hc = HingeAnalyzer(graph=graph,
                           pdb_name=pdb_name,
                           destination_output_path=destination_output_path,
                           actual_residue_number_map=actual_residue_number_map)
        nums_cluster = None
        if _CLUSTER_NUMBER in calculation_options and calculation_options[_CLUSTER_NUMBER][IS_CHECKED]:
            nums_cluster = calculation_options[_CLUSTER_NUMBER][VALUE]
        hc.compute_hinge_residues_with_sign(num_modes=nums_cluster)
        return hc

    @log_with_stars("Centrality PDB Mapper")
    @log_time("Centrality PDB Mapper")
    def _process_centrality_pdb_mapper(self, graph=None, pdb_name: str = None, destination_output_path: str = None,
                                       calculation_options: dict = None) -> CentralityPdbMapper:
        """Process creating PDBs to visualize by mapping centrality scores into the given PDB structure."""
        cpm = CentralityPdbMapper(graph=graph, pdb_name=pdb_name, destination_output_path=destination_output_path,
                                  calculation_options=calculation_options)
        cpm.process()
        return cpm

    @log_with_stars("Quantile Analyzer")
    @log_time("Quantile Analyzer")
    def _process_quantiles(self, ca: CentralityAnalyzer = None, use_parallel: bool = False):
        """Process quantiles to calculate the quantile values for the three centrality types."""
        ca.calculate_quantiles(use_parallel=use_parallel)

    @log_time("Quantile Analyzer")
    def _process_quantile_centrality(self, high_percentage_dict=None,
                                     centrality_type: CentralityType = CentralityType.BET,
                                     destination_output_path=None) -> QuantileAnalyzer:
        """Process quantile calculations to find common residues with high centrality scores among the given PDBs"""
        qa = QuantileAnalyzer(high_percentage_dict=high_percentage_dict, centrality_type=centrality_type,
                              destination_output_path=destination_output_path)
        qa.run_analysis()

        return qa

    @log_with_stars("Processing Quantile Outputs")
    @log_time("Processing Quantile Outputs")
    def _process_quantile_output(self, bet_high_percentage_dict=None, clos_high_percentage_dict=None,
                                 deg_high_percentage_dict=None, destination_output_path=None, calculation_options=None):
        if calculation_options[CentralityType.BET.display_name][IS_CHECKED] and bet_high_percentage_dict:
            logging.info(f"bet_high_percentage_dict: {bet_high_percentage_dict}")
            self._process_quantile_centrality(high_percentage_dict=bet_high_percentage_dict,
                                              centrality_type=CentralityType.BET,
                                              destination_output_path=destination_output_path)

        if calculation_options[CentralityType.CLOS.display_name][IS_CHECKED] and clos_high_percentage_dict:
            logging.info(f"clos_high_percentage_dict: {clos_high_percentage_dict}")
            self._process_quantile_centrality(high_percentage_dict=clos_high_percentage_dict,
                                              centrality_type=CentralityType.CLOS,
                                              destination_output_path=destination_output_path)
        if calculation_options[CentralityType.DEG.display_name][IS_CHECKED] and deg_high_percentage_dict:
            logging.info(f"deg_high_percentage_dict: {deg_high_percentage_dict}")
            self._process_quantile_centrality(high_percentage_dict=deg_high_percentage_dict,
                                              centrality_type=CentralityType.DEG,
                                              destination_output_path=destination_output_path)

    @log_time("Extracting PDB IDs to be processed.")
    def _extract_output_path_pdb_ids(self, output_path, input_path, pdb_ids, trajectory_file) -> list:
        """Extracts PDB IDs from the given input option."""
        output_path_pdb_ids = []
        if input_path is not None and input_path != "":
            logging.info(f"input_path: {input_path}")
            input_pdb_files = [pdb for pdb in glob.glob(os.path.join(input_path, "*.pdb"))]
            for input_pdb in input_pdb_files:
                input_pdb_path = Path(input_pdb)
                pdb_code = input_pdb_path.stem
                destination_path = os.path.join(output_path, pdb_code)
                create_folder_not_exists(destination_path)
                shutil.copy(input_pdb, destination_path)
                output_path_pdb_ids.append(os.path.join(destination_path, input_pdb_path.name))
        elif pdb_ids is not None and pdb_ids:
            logging.info(f"fetch_pdb_ids: {self.pdb_ids}")
            fetch_pdb_ids = [x.strip() for x in self.pdb_ids.split(',')]
            output_path_pdb_ids = self._fetch_pdb_ids(fetch_pdb_ids, output_path)
        elif trajectory_file is not None and trajectory_file != "":
            logging.info(f"trajectory_file: {trajectory_file}")
            trajectory = Trajectory(input_trajectory_file=self.trajectory_file, output_path=output_path,
                                    stride=self.stride)
            output_path_pdb_ids = trajectory.parse_to_pdb()
        else:
            logging.warning(f"There is no proper input selection to be processed! Please check your input!")
            raise Exception("Please Select Only One Input Option!!!")

        return output_path_pdb_ids

    @staticmethod
    @log_with_stars("Process Summary")
    def _print_summary(not_processed_pdb_ids, output_path, success_pdb_ids, total_amino_acids_map,
                       total_start_time):
        total_pdb_ids = len(success_pdb_ids) + len(not_processed_pdb_ids)
        logging.info(
            f"PDB processing summary: "
            f"{len(success_pdb_ids)} out of {total_pdb_ids} PDB structures were successfully processed; "
            f"{len(not_processed_pdb_ids)} structures were skipped because edge weights could not be constructed."
        )
        if not_processed_pdb_ids:
            logging.info(
                f"{len(not_processed_pdb_ids)} PDB structures were not processed: "
                f"{', '.join(not_processed_pdb_ids)}"
            )
        logging.info(
            f"Total PDBs: {len(total_amino_acids_map)}, Total amino acids: {sum(total_amino_acids_map.values())}")
        total_end_time = datetime.strptime(datetime.now().strftime("%H:%M:%S"), "%H:%M:%S")
        delta = total_end_time - total_start_time
        total_seconds = int(delta.total_seconds())
        hours, remainder = divmod(total_seconds, 3600)
        minutes, seconds = divmod(remainder, 60)
        logging.info(f"Total elapsed duration is {delta.total_seconds()} seconds ({hours}h {minutes}m {seconds}s).")
        logging_config.save_logs(output_dir=output_path)

    @log_with_stars("[RINProcess] Module")
    @log_time("RIN Process")
    def start_process(self) -> bool:
        # clear the debug.log file before starting the processes.
        logging_config.clear_logs()
        output_path = self.output_path
        logging.info(f"output_path: {output_path}")

        output_path_pdb_ids = self._extract_output_path_pdb_ids(output_path=output_path,
                                                                input_path=self.input_path,
                                                                pdb_ids=self.pdb_ids,
                                                                trajectory_file=self.trajectory_file)
        bet_high_percentage_dict = {}
        clos_high_percentage_dict = {}
        deg_high_percentage_dict = {}
        total_amino_acids_map = {}
        success_pdb_ids = []
        not_processed_pdb_ids = []

        if output_path_pdb_ids:
            logging.info(f"Processing PDB files: {output_path_pdb_ids}")
            total_start_time = datetime.strptime(datetime.now().strftime("%H:%M:%S"), "%H:%M:%S")
            for pdb_file_path in output_path_pdb_ids:
                pdb_name = Path(str(pdb_file_path)).stem

                start_time = datetime.strptime(datetime.now().strftime("%H:%M:%S"), "%H:%M:%S")

                log_util.log_star_message(pdb_name.upper())
                logging.info(f'Start time: {start_time.time()}')

                residue_network_builder, is_success = self._process_residue_network_builder(pdb_name=pdb_name,
                                                                                            pdb_path=pdb_file_path,
                                                                                            use_preprocess=True,
                                                                                            destination_output_path=output_path,
                                                                                            calculation_options=self.calculation_options)

                if not is_success:
                    logging.info(
                        f"{pdb_name}: No residue–residue interactions were detected "
                        f"within the specified distance cutoff "
                        f"(≤ {residue_network_builder.cutoff} Å). "
                        f"Edge weights could not be constructed."
                    )
                    logging.warning(
                        f"{pdb_name}: No residue–residue interactions were detected "
                        f"within the specified distance cutoff "
                        f"(≤ {residue_network_builder.cutoff} Å). "
                        f"Edge weights could not be constructed."
                    )

                    not_processed_pdb_ids.append(pdb_name)
                    continue

                number_of_amino_acid = residue_network_builder.get_number_of_amino_acid()
                total_amino_acids_map[pdb_name] = number_of_amino_acid

                ca = self._process_centrality_analyzer(pdb_name=pdb_name,
                                                       number_of_amino_acid=number_of_amino_acid,
                                                       destination_output_path=output_path,
                                                       calculation_options=self.calculation_options,
                                                       actual_residue_number_map=residue_network_builder.get_actual_residue_number_map())

                bet_high_percentage_filename = f'{pdb_name}_{HIGH_PERCENTAGE_TEMPLATE.format(type=CentralityType.BET.display_name)}'
                clos_high_percentage_filename = f'{pdb_name}_{HIGH_PERCENTAGE_TEMPLATE.format(type=CentralityType.CLOS.display_name)}'
                deg_high_percentage_filename = f'{pdb_name}_{HIGH_PERCENTAGE_TEMPLATE.format(type=CentralityType.DEG.display_name)}'

                bet_high_percentage_dict[pdb_name] = os.path.join(output_path, pdb_name, bet_high_percentage_filename)
                clos_high_percentage_dict[pdb_name] = os.path.join(output_path, pdb_name, clos_high_percentage_filename)
                deg_high_percentage_dict[pdb_name] = os.path.join(output_path, pdb_name, deg_high_percentage_filename)

                cpm = self._process_centrality_pdb_mapper(graph=ca.graph,
                                                          pdb_name=pdb_name,
                                                          destination_output_path=output_path,
                                                          calculation_options=self.calculation_options)

                self._process_quantiles(ca=ca, use_parallel=True)

                hc = self._process_hinge_analyzer(graph=ca.graph,
                                                  pdb_name=pdb_name,
                                                  destination_output_path=output_path,
                                                  actual_residue_number_map=residue_network_builder.get_actual_residue_number_map(),
                                                  calculation_options=self.calculation_options)

                end_time = datetime.strptime(datetime.now().strftime("%H:%M:%S"), "%H:%M:%S")
                logging.info(f'End time: {end_time.time()}')
                delta = end_time - start_time
                logging.info(f"Elapsed duration for {pdb_name.upper()} is {delta.total_seconds()} seconds")
                log_util.log_star_message(pdb_name.upper())
                success_pdb_ids.append(pdb_name)

            if len(success_pdb_ids) > 1:
                self._process_quantile_output(bet_high_percentage_dict=bet_high_percentage_dict,
                                              clos_high_percentage_dict=clos_high_percentage_dict,
                                              deg_high_percentage_dict=deg_high_percentage_dict,
                                              destination_output_path=output_path,
                                              calculation_options=self.calculation_options)

            self._print_summary(not_processed_pdb_ids, output_path, success_pdb_ids, total_amino_acids_map,
                                total_start_time)

            return True
        else:
            logging.warning(f"There is no PDB file to be processed! Please check your input!")
            raise Exception("NO PDB FILES TO BE PROCESSED!")


def _load_dict_file(file_path: Path):
    if not file_path.exists():
        print(f"Warning: ligand_dict file not found at '{file_path}'. Skipping.")
        return None
    try:
        with file_path.open('r') as f:
            return json.load(f)
    except json.JSONDecodeError as e:
        print(f"Error decoding JSON: {e}")
        return None


def build_parser():
    parser = argparse.ArgumentParser(
        description="RinPy: Residue Interaction Network construction and analysis from protein structures and trajectories."
    )
    parser.add_argument("--output_path",
                        type=str,
                        help="Specify output path which will store the results.",
                        required=True)
    group = parser.add_mutually_exclusive_group(required=True)
    group.add_argument(
        "--input_path",
        type=str,
        help="Specify the path of PDB files which will be analyzed."
    )
    group.add_argument(
        "--pdb_ids",
        type=str,
        help="Load a PDB file from the Protein Data Bank via its unique 4-letter ID, e.g., '6epl, 6epm, 6epn'"
    )
    group.add_argument(
        "--trajectory_file",
        type=str,
        help="Specify the trajectory file to be analyzed."
    )
    parser.add_argument("--stride",
                        type=int,
                        required=False,
                        help="Parsing the trajectory file every this stride number.",
                        default=1)
    parser.add_argument("--calculation_option_file",
                        type=Path,
                        required=True,
                        help="Required: Path to JSON file containing the calculation options in the dictionary.",
                        default=None)
    parser.add_argument("--ligand_dict_file",
                        type=Path,
                        required=False,
                        help="Optional: Path to JSON file containing the ligand dictionary.",
                        default=None)
    parser.add_argument("--num_workers",
                        type=int,
                        default=None,
                        help="Number of CPU worker processes to use (default: use all detected CPU cores.)")
    args = parser.parse_args()
    return args


def main():
    args = build_parser()

    ligand_dict = None

    if args and args.ligand_dict_file is not None:
        ligand_dict = _load_dict_file(args.ligand_dict_file)

    calculation_options = _load_dict_file(args.calculation_option_file)

    rin = RINProcess(input_path=args.input_path,
                     output_path=args.output_path,
                     pdb_ids=args.pdb_ids,
                     ligand_dict=ligand_dict,
                     calculation_options=calculation_options,
                     trajectory_file=args.trajectory_file,
                     stride=args.stride,
                     num_workers=args.num_workers)

    rin.start_process()


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
"""
    __author__ = 'Zehra Sarica'
    __email__ = ['sarica16@itu.edu.tr','zehraacar559@gmail.com']
"""

import logging
import os
from importlib import resources

import matplotlib.font_manager as fm
import matplotlib.pyplot as plt
import plotly.graph_objects as go
import plotly.io as pio

_fonts_loaded = False


def load_fonts_once():
    """ Load packaged Times New Roman fonts exactly once per interpreter. """
    global _fonts_loaded
    if _fonts_loaded:
        return

    try:
        with resources.as_file(
                resources.files("rinpy") / "fonts"
        ) as font_dir:
            for font_file in os.listdir(font_dir):
                if font_file.endswith(".ttf"):
                    fm.fontManager.addfont(
                        os.path.join(font_dir, font_file)
                    )

        plt.rcParams["font.family"] = "Times New Roman"

        pio.templates["custom_times"] = go.layout.Template(
            layout=go.Layout(font=dict(family="Times New Roman"))
        )
        pio.templates.default = "custom_times"
        logging.info(f"Active font: {plt.rcParams['font.family'][0]}")

    except Exception as e:
        plt.rcParams["font.family"] = "DejaVu Serif"
        logging.warning(f"Could not load Times font ({e}). Using default {plt.rcParams['font.family'][0]}.")

    _fonts_loaded = True



# -*- coding: utf-8 -*-
"""
    __author__ = 'Zehra Sarica'
    __email__ = ['sarica16@itu.edu.tr','zehraacar559@gmail.com']
"""

import logging
import os

import matplotlib as mpl
import matplotlib.pyplot as plt
import networkx as nx
import numpy as np
import plotly.graph_objects as go
import seaborn as sns
from matplotlib import ticker
from scipy.sparse.linalg import eigsh

from rinpy import utils
from rinpy.constants import CHAIN_ID, RESIDUE_NAME, SEGMENT_ID
from rinpy.style_config import FONT_STYLES, FONT_FAMILY, EDGE_COLOR, COLOR_PALETTE

X_COORD_LABEL = "X Coordinate"
Y_COORD_LABEL = "Y Coordinate"
Z_COORD_LABEL = "Z Coordinate"


class GraphPlotter:
    def __init__(self, pdb_name, destination_output_path, actual_residue_number_map):
        self.pdb_name = pdb_name
        self.destination_output_path = destination_output_path
        self.actual_residue_number_map = actual_residue_number_map
        mpl.rcParams['font.family'] = FONT_FAMILY

    def get_full_save_path(self, filename: str, extension: str = "png"):
        return os.path.join(self.destination_output_path, self.pdb_name, f"{self.pdb_name}_{filename}.{extension}")

    @staticmethod
    def _apply_colorbar_style(cbar, cbar_label=None):
        """Apply consistent style to colorbar."""
        cbar.ax.tick_params(labelsize=FONT_STYLES["colorbar"]["tick_params"]["labelsize"])
        if cbar_label is not None:
            cbar.set_label(
                "Colorbar Label",
                fontsize=FONT_STYLES["colorbar"]["label"]["fontsize"],
                fontfamily=FONT_STYLES["colorbar"]["label"]["fontfamily"]
            )

    def plot_graph_2d(self, graph, filename="graph", centrality_scores: dict = None, title=None, cmap='jet',
                      node_size=40, alpha=0.8, font_size=6, font_color='black', equal_axis=False):
        """ Creates a 2D plot of a graph, optionally with centrality-based coloring of nodes.

           Parameters:
           - graph: The networkx graph object to be plotted.
           - filename: The name of the file to save the plot.
           - centrality_scores: Dictionary with node IDs as keys and centrality scores as values.
           - title: The title to be displayed on the plot.
           - cmap: Colormap for node color when centrality_scores is provided.
           - node_size: Size of nodes in the plot.
           - alpha: Transparency level for nodes and edges.
           - font_size: Font size for node labels.
           - font_color: Font color for node labels.
           - equal_axis: If True, sets the axes to be equal in scale.
       """

        full_path = self.get_full_save_path(filename=filename)

        plt.figure()

        pos = {n: (graph.nodes[n]['x'], graph.nodes[n]['y']) for n in graph.nodes}

        node_colors = [centrality_scores.get(n, 0) for n in graph.nodes] if centrality_scores else None

        nodes = nx.draw_networkx_nodes(graph,
                                       pos,
                                       node_color=node_colors,
                                       cmap=plt.get_cmap(cmap) if centrality_scores is not None else None,
                                       node_size=node_size,
                                       alpha=alpha)
        nx.draw_networkx_labels(graph,
                                pos,
                                labels={n: str(n) for n in graph.nodes()},
                                font_color=font_color,
                                font_size=font_size)

        nx.draw_networkx_edges(graph,
                               pos,
                               edge_color='gray',
                               alpha=alpha,
                               width=0.5)

        if node_colors:
            cbar = plt.colorbar(nodes)
            self._apply_colorbar_style(cbar)

        ax = plt.gca()
        ax.set_xlabel(X_COORD_LABEL, fontdict=FONT_STYLES["xlabel"])
        ax.set_ylabel(Y_COORD_LABEL, fontdict=FONT_STYLES["ylabel"])
        ax.tick_params(left=True, bottom=True, labelleft=True, labelbottom=True,
                       labelsize=FONT_STYLES["tick"]["labelsize"])
        ax.grid(True, alpha=0.3)

        if title is not None:
            plt.title(f"{title} - {self.pdb_name}", fontdict=FONT_STYLES["title"])

        if equal_axis:
            ax.axis('equal')

        plt.tight_layout()
        plt.savefig(full_path, dpi=300)
        plt.close()

    def plot_graph_3d(self, graph, filename="graph_3D", title="3D Graph Visualization"):
        """ Creates a 3D plot of a graph using Plotly and saves it as an interactive HTML file. """

        full_path = self.get_full_save_path(filename=filename, extension="html")

        pos = {n: (graph.nodes[n]['x'], graph.nodes[n]['y'], graph.nodes[n]['z']) for n in graph.nodes}
        edge_x, edge_y, edge_z = [], [], []
        for u, v in graph.edges():
            x0, y0, z0 = pos[u]
            x1, y1, z1 = pos[v]
            edge_x += [x0, x1, None]
            edge_y += [y0, y1, None]
            edge_z += [z0, z1, None]

        edge_trace = go.Scatter3d(
            x=edge_x,
            y=edge_y,
            z=edge_z,
            mode='lines',
            line=dict(width=1, color='gray'),
            hoverinfo='none',
            showlegend=True,
            name="Edge"
        )

        chain_to_nodes = {}
        for node_id in graph.nodes():
            current_node = graph.nodes[node_id]
            chain_id = current_node['chain_id']
            chain_to_nodes.setdefault(chain_id, []).append(node_id)

        node_traces = []
        for i, (chain_id, node_ids) in enumerate(chain_to_nodes.items()):
            color = COLOR_PALETTE[i] if i < len(COLOR_PALETTE) else None
            node_x = [graph.nodes[node_id]['x'] for node_id in node_ids]
            node_y = [graph.nodes[node_id]['y'] for node_id in node_ids]
            node_z = [graph.nodes[node_id]['z'] for node_id in node_ids]
            hovertext = []
            labels = []
            for node_id in node_ids:
                node = graph.nodes[node_id]
                seg_text = f"Segment ID: {node[SEGMENT_ID]}<br>" if node[SEGMENT_ID] not in ['', "''", None] else ''
                text = (
                    f"Chain: {node[CHAIN_ID]}<br>"
                    f"Residue Number: {utils.get_residue_id(node)}<br>"
                    f"Residue Name: {node[RESIDUE_NAME]}<br>"
                    f"{seg_text}"
                    f"X: {node['x']:.3f}<br>"
                    f"Y: {node['y']:.3f}<br>"
                    f"Z: {node['z']:.3f}"
                )
                hovertext.append(text)

                res_id = utils.get_residue_id(node)
                seg_id = node.get(SEGMENT_ID)
                if seg_id not in [None, '', "''"]:
                    res_id = f"{res_id} ({seg_id})"
                labels.append(res_id)

            trace = go.Scatter3d(
                x=node_x, y=node_y, z=node_z,
                mode='markers+text',
                marker=dict(size=6, color=color),
                text=labels,
                textposition='top center',
                hoverinfo='text',
                hovertext=hovertext,
                name=f"Chain {chain_id}",
                legendgroup=f"Chain {chain_id}",
                showlegend=True,
                hoverlabel=dict(
                    font=dict(color='white')
                )
            )
            node_traces.append(trace)

        fig = go.Figure(data=[edge_trace] + node_traces)
        fig.update_layout(
            title=f"{title} - {self.pdb_name}",
            showlegend=True,
            margin=dict(l=0, r=0, b=0, t=40),
            scene=dict(
                xaxis=dict(title=X_COORD_LABEL),
                yaxis=dict(title=Y_COORD_LABEL),
                zaxis=dict(title=Z_COORD_LABEL),
                aspectmode='data'
            ),
            template='plotly'
        )

        fig.write_html(full_path)

    def plot_graph_2d_degree(self, graph, filename="degree_graph", title=None, node_circle_size=80, equal_axis=False,
                             font_size=6, font_color='white'):

        full_path = self.get_full_save_path(filename=filename, extension="png")
        nodes = sorted(graph.nodes())
        laplacian = nx.laplacian_matrix(graph, nodelist=nodes).astype(float)
        eigenvalues, eigenvectors = eigsh(laplacian, k=4, which="SM")
        w = eigenvectors[:, 0].real
        pos = {residue_index: (graph.nodes[residue_index]['x'], graph.nodes[residue_index]['y']) for residue_index in
               graph.nodes}
        plt.figure()
        node_color = ['r' if w[i] >= 0 else 'k' for i in range(len(w))]

        nx.draw_networkx_nodes(graph,
                               pos,
                               node_color=node_color,
                               node_size=node_circle_size)
        nx.draw_networkx_labels(graph,
                                pos,
                                labels={n: str(n) for n in graph.nodes()},
                                font_color=font_color,
                                font_size=font_size)
        nx.draw_networkx_edges(graph,
                               pos,
                               edge_color=EDGE_COLOR,
                               width=0.5)

        ax = plt.gca()
        ax.set_xlabel(X_COORD_LABEL, fontdict=FONT_STYLES["xlabel"])
        ax.set_ylabel(Y_COORD_LABEL, fontdict=FONT_STYLES["ylabel"])
        ax.tick_params(left=True, bottom=True, labelleft=True, labelbottom=True,
                       labelsize=FONT_STYLES["tick"]["labelsize"])
        ax.grid(True, alpha=0.3)

        if title is not None:
            plt.title(f"{title} - {self.pdb_name}", fontdict=FONT_STYLES["title"])

        if equal_axis:
            ax.axis('equal')

        plt.tight_layout()
        plt.savefig(full_path, dpi=300)
        plt.close()

    def plot_graph_2d_degree_with_coords(self, graph, filename="degree_coords_graph", cmap='jet', title=None,
                                         equal_axis=False, font_size=6, font_color='black'):
        full_path = self.get_full_save_path(filename=filename, extension="png")
        pos = {n: (graph.nodes[n]['x'], graph.nodes[n]['y']) for n in graph.nodes}

        plt.figure()

        deg = np.array([degree for node, degree in graph.degree()])

        node_sizes = 20 * np.sqrt(deg - np.min(deg) + 0.2)

        node_colors = deg

        nodes = nx.draw_networkx_nodes(graph,
                                       pos,
                                       cmap=cmap,
                                       node_color=node_colors,
                                       node_size=node_sizes)
        nx.draw_networkx_labels(graph,
                                pos,
                                labels={n: str(n) for n in graph.nodes()},
                                font_color=font_color,
                                font_size=font_size)
        nx.draw_networkx_edges(graph,
                               pos,
                               edge_color=EDGE_COLOR,
                               width=0.5,
                               alpha=0.5)
        cbar = plt.colorbar(nodes)
        self._apply_colorbar_style(cbar)

        ax = plt.gca()
        ax.tick_params(left=True, bottom=True, labelleft=True, labelbottom=True,
                       labelsize=FONT_STYLES["tick"]["labelsize"])
        ax.axis('on')

        ax.set_xlabel(X_COORD_LABEL, fontdict=FONT_STYLES["xlabel"])
        ax.set_ylabel(Y_COORD_LABEL, fontdict=FONT_STYLES["ylabel"])

        if title is not None:
            plt.title(f"{title} - {self.pdb_name}", fontdict=FONT_STYLES["title"])

        if equal_axis:
            ax.axis('equal')

        plt.tight_layout()
        plt.savefig(full_path, dpi=300)
        plt.close()

    def plot_figures(self, scores_dict: dict, centrality_type_name: str, use_decimal: bool = True) -> None:
        """ Plots centrality scores for residues using actual residue numbers on the x-axis.

        Parameters:
        - scores_dict (dict): Mapping of residue index to centrality score.
        - centrality_type_name (str): Type of centrality (e.g., 'betweenness', 'closeness').
        - use_decimal (bool): Whether to format y-axis with decimal values.
        """
        full_path = self.get_full_save_path(filename=f"{centrality_type_name}_scores", extension="png")
        sorted_data = dict(sorted(scores_dict.items()))
        keys = list(sorted_data.keys())
        values = list(sorted_data.values())

        plt.figure(figsize=(10, 6))
        plt.plot(keys, values, linestyle='-')

        plt.ylim([0, max(values) * 1.05])

        if use_decimal:
            plt.gca().yaxis.set_major_formatter(plt.FuncFormatter(lambda x, _: f'{x:.2f}'))
        else:
            plt.gca().yaxis.set_major_formatter(plt.FuncFormatter(lambda x, _: f'{int(x)}'))

        margin = 0.05 * (max(keys) - min(keys))
        plt.xlim([min(keys) - margin, max(keys) + margin])

        ax = plt.gca()
        ax.xaxis.set_major_locator(ticker.MaxNLocator(integer=True))
        auto_x_ticks = ax.get_xticks().astype(int)
        visible_x_ticks = [tick for tick in auto_x_ticks if min(keys) <= tick <= max(keys)]
        if min(keys) not in visible_x_ticks:
            visible_x_ticks.insert(0, min(keys))
        if max(keys) not in visible_x_ticks:
            visible_x_ticks.append(max(keys))

        xtick_labels = [
            f"{i} ({self.actual_residue_number_map[i][1]}, {utils.get_residue_id_by_tuple(self.actual_residue_number_map[i])})"
            if i in self.actual_residue_number_map else i
            for i in visible_x_ticks
        ]
        plt.xticks(ticks=visible_x_ticks, labels=xtick_labels, rotation=45, fontsize=FONT_STYLES['xtick']['labelsize'],
                   fontname=FONT_FAMILY)
        plt.yticks(fontsize=FONT_STYLES['ytick']['labelsize'], fontname=FONT_FAMILY)

        has_any_seg_id = any(
            seg_id not in [None, '', "''"]
            for res_info in self.actual_residue_number_map.values()
            for _, _, _, _, seg_id, _ in [res_info]
        )
        if has_any_seg_id:
            plt.xlabel('Residue Index (Chain ID, Residue Number, Segment ID)', fontdict=FONT_STYLES["xlabel"])
        else:
            plt.xlabel('Residue Index (Chain ID, Residue Number)', fontdict=FONT_STYLES["xlabel"])
        plt.ylabel(f'{centrality_type_name.capitalize()} Centrality Score', fontdict=FONT_STYLES["ylabel"])

        plt.title(f"PDB ID: {self.pdb_name}", fontdict=FONT_STYLES["title"])

        plt.grid(alpha=0.5)
        plt.tight_layout()

        plt.savefig(full_path, dpi=300)
        logging.info(f"{centrality_type_name.capitalize()} scores have been saved to {full_path}.")

    def plot_histogram(self, scores_dict: dict, centrality_type_name: str) -> None:
        """ Plots and saves a histogram of centrality scores.

            Parameters:
            -----------
            scores_dict : dict
                A dictionary where keys are residue indices and values are centrality scores.

            centrality_type_name : str
                The name of the centrality type (e.g., 'closeness', 'betweenness') to be used in axis labels and file naming.

            Description:
            ------------
            This method generates a histogram showing the frequency distribution of centrality scores
            provided in `scores_dict`. The histogram is saved as a PNG file with a filename based on
            `centrality_type_name`. Axis labels, tick formatting, and layout are adjusted for clarity.
        """

        full_path = self.get_full_save_path(filename=f"{centrality_type_name}_scores_histogram", extension="png")

        sorted_data = dict(sorted(scores_dict.items()))

        values = list(sorted_data.values())

        plt.figure(figsize=(10, 6))

        hist_plot = sns.histplot(values, kde=False, bins=10)

        plt.xlabel(f'{centrality_type_name.capitalize()} Centrality', fontdict=FONT_STYLES["xlabel"])
        plt.ylabel('Frequency', fontdict=FONT_STYLES["xlabel"])

        plt.xlim(left=0, right=np.ceil(max(values) * 10) / 10)
        max_y_value = hist_plot.get_ylim()[1]
        plt.ylim(bottom=0, top=np.ceil(max_y_value))
        plt.title(f"PDB ID: {self.pdb_name}", fontdict=FONT_STYLES["title"])
        plt.tight_layout()
        plt.savefig(full_path, dpi=300)
        plt.close()
        logging.info(f"{centrality_type_name.capitalize()} scores has been saved to {full_path}.")


def main():
    pass


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
"""
    __author__ = 'Zehra Sarica'
    __email__ = ['sarica16@itu.edu.tr','zehraacar559@gmail.com']
"""

import logging
import pathlib
import shutil

RINPY_DEFAULT_FOLDER = ".rinpy"
DEBUG_LOG = "debug.log"
PROCESS_LOG = "process.log"
LOG_TXT = "log.txt"
UTF_8_ENCODING = 'utf-8'


class InfoOnlyFilter(logging.Filter):
    """Filter only INFO messages to be shown in the log files."""

    def filter(self, record):
        return record.levelno == logging.INFO


def get_rinpy_log_dir():
    """Get the rinpy log directory path."""
    home_dir = pathlib.Path.home()
    rinpy_dir = home_dir / RINPY_DEFAULT_FOLDER
    rinpy_dir.mkdir(parents=True, exist_ok=True)
    return rinpy_dir


log_dir = get_rinpy_log_dir()

info_log_file = log_dir / PROCESS_LOG
all_log_file = log_dir / DEBUG_LOG

formatter = logging.Formatter('%(asctime)s.%(msecs)03d %(levelname)s - %(message)s', datefmt='%Y-%m-%d %H:%M:%S')

info_handler = logging.FileHandler(info_log_file, encoding=UTF_8_ENCODING)
info_handler.setLevel(logging.INFO)
info_handler.addFilter(InfoOnlyFilter())
info_handler.setFormatter(formatter)

all_handler = logging.FileHandler(all_log_file, encoding=UTF_8_ENCODING)
all_handler.setLevel(logging.DEBUG)
all_handler.setFormatter(formatter)

stream_handler = logging.StreamHandler()
stream_handler.setLevel(logging.DEBUG)
stream_handler.setFormatter(formatter)

logging.basicConfig(
    level=logging.DEBUG,
    handlers=[info_handler, all_handler, stream_handler]
)

logging.getLogger("matplotlib").setLevel(logging.WARNING)
logging.getLogger("urllib3").setLevel(logging.WARNING)


def clear_logs():
    """Clear the content of log files for each run."""
    for file in [info_log_file, all_log_file]:
        try:
            if file.exists():
                with open(file, "r+", encoding=UTF_8_ENCODING) as f:
                    f.truncate()
        except Exception as e:
            logging.warning(f"Log file cannot be cleared: {file} ({e})")


def save_logs(output_dir: str):
    """ Save .rinpy/process.log to the given output directory as log.txt.
    If .rinpy/process.log does not exist, log a warning.
    """
    process_log_file = log_dir / PROCESS_LOG  # .rinpy/process.log path

    try:
        if not process_log_file.exists():
            logging.warning(f"Process log not found at: {process_log_file}")
            return

        destination = pathlib.Path(output_dir)
        destination.mkdir(parents=True, exist_ok=True)

        output_file = destination / LOG_TXT
        shutil.copyfile(process_log_file, output_file)

        logging.info(f"Process log saved to: {output_file}")

    except Exception as e:
        logging.error(f"Failed to save process log to {output_dir}: {e}")



# -*- coding: utf-8 -*-
"""
    __author__ = 'Zehra Sarica'
    __email__ = ['sarica16@itu.edu.tr','zehraacar559@gmail.com']
"""
import logging
import multiprocessing as mp
import os

import matplotlib
import networkx as nx
import numpy as np

# Force non-interactive backend before any matplotlib import.
# Prevents Tkinter RuntimeError in multiprocessing workers on Windows.
if matplotlib.get_backend() != "Agg":
    matplotlib.use("Agg")

from rinpy import utils
from rinpy.constants import CENTRALITY_PDB_TEMPLATE, HIGH_PERCENTAGE_TEMPLATE, RESIDUE_NUMBER, VALUE, \
    CENTRALITY_CSV_TEMPLATE, RESIDUE_EDGES_FILE, RESIDUE_AVERAGE_COORDINATES_FILE, IS_CHECKED, \
    RESIDUE_NAME, CHAIN_ID, INSERTION, X_COORD, Z_COORD, Y_COORD, RESIDUE_INDEX, TOP_PERCENTAGE_TILE, SOURCE_RESIDUE, \
    TARGET_RESIDUE, X, Y, Z, GRAPH_NAME, ATOM_NUMBER, CENTRALITY_SCORE, SEGMENT_ID, RECORD_NAME, HETATM, APO, HOLO
from rinpy.graph_plotter import GraphPlotter
from rinpy.log_util import log_details, log_time
from rinpy.pymol_utils import PymolUtils
from rinpy.utils import CentralityType

CENTRALITY_TYPE_KEY = 'centrality_type'
QUANTILE_PERCENTAGE = "quantile_percentage"
OUTFILE = "outfile"
PDB_FILE = "pdb_file"
CSV_FILE = "csv_file"


class CentralityAnalyzer:
    """ The Centrality Analyzer module generates output data for constructing residue interaction networks based on
        affinity calculations between α-carbon atoms. It computes atom-to-atom distances between source and target
        residues to identify potential contacts.

        Attributes
        ----------
        pdb_name : str, default: None
            PDB ID which must be same with file name, such as 4OBE (4OBE.pdb).

        number_of_amino_acid : int, default: None
            The number of residue in the given PDB file.

        destination_output_path : str, default: output, this path must be provided.
            Location of the output results in which all files for each PDB file process are stored in this path.

        calculation_options: dict, default: None
            The calculation options as the input for the whole process.
            Example:
            {
                'remove_hydrogen': {
                    'is_checked': True,
                    'value': 0
                },
                'betweenness': {
                    'is_checked': True,
                    'value': 5
                },
                'closeness': {
                    'is_checked': False,
                    'value': 10
                },
                'degree': {
                    'is_checked': False,
                    'value': 10
                },
                'cluster_number': {
                    'is_checked': None,
                    'value': 4
                },
                'cutoff': {
                    'is_checked': True,
                    'value': 4.5
                },
            }
        actual_residue_number_map: dict[int, tuple[str, str, int, str, str, str]], default: None
        """

    def __init__(self, pdb_name, number_of_amino_acid, destination_output_path=None, calculation_options=None,
                 actual_residue_number_map: dict[int, tuple[str, str, int, str, str, str]] = None):
        if destination_output_path is None:
            raise ValueError('You must provide an output path to proceed.')
        self.pdb_name = pdb_name
        self.calculation_options = calculation_options
        self.number_of_amino_acid = number_of_amino_acid
        logging.info(f"Number of amino acid: {self.number_of_amino_acid}")
        self.destination_output_path = destination_output_path
        utils.create_folder_not_exists(os.path.join(self.destination_output_path))
        self.actual_residue_number_map = actual_residue_number_map
        self.edges = None
        self.weights = None
        self.number_of_nodes = 0
        self.graph = None
        self.graph_plotter = GraphPlotter(pdb_name=self.pdb_name,
                                          destination_output_path=self.destination_output_path,
                                          actual_residue_number_map=self.actual_residue_number_map)

        self.pymol_utils = PymolUtils(pdb_name=pdb_name,
                                      destination_output_path=destination_output_path,
                                      actual_residue_number_map=actual_residue_number_map)

    def _compute_high_percentage_quantile(self, centrality_type: CentralityType = CentralityType.DEG,
                                          centrality_csv_file: str = CENTRALITY_CSV_TEMPLATE.format(
                                              type=CentralityType.BET.display_name),
                                          high_percentage_file: str = HIGH_PERCENTAGE_TEMPLATE.format(
                                              type=CentralityType.BET.display_name),
                                          quantile_percentage=5):
        """ Creates quantile files ({pdb_id}_betweenness_high_percentage_residues.csv,
        {pdb_id}_closeness_high_percentage_residues.csv, and {pdb_id}_degree_high_percentage_residues.csv)
        for the given input file ({pdb_id}_centrality_betweenness.csv or {pdb_id}_centrality_closeness.csv or
        {pdb_id}_centrality_degree.csv) and its corresponding hub file with the given quantile percentage.
        Also, it generates PyMOL query strings to visualizes hub residues on the protein 3D structure.

        Parameters:
        -----------
        centrality_type : CentralityType, default CentralityType.DEG
            The centrality type for centrality analysis.
            strength.
        centrality_csv_file : str, default {PDB_ID}_centrality_betweennes.csv
            The PDB file contains the calculated centrality score (betweenness, closeness, and degree) in
            the b-factor column of the PDB.
        high_percentage_file : str, default betweennes_high_percentage_residues.csv
            The file will store the top quantile scores of the given input.
        quantile_percentage : int, default 5
            The threshold value to obtain the top quantile of the given percentage.
        """

        log_detail_name = f'{centrality_type.display_name} - {self.pdb_name}'
        logging.info(f'Given quantile percentage for {log_detail_name} : {quantile_percentage}')

        columns = [ATOM_NUMBER, CENTRALITY_SCORE, RESIDUE_NAME, CHAIN_ID, RESIDUE_NUMBER, INSERTION, SEGMENT_ID]
        dtypes = {ATOM_NUMBER: int, CENTRALITY_SCORE: float, RESIDUE_NAME: str, CHAIN_ID: str, RESIDUE_NUMBER: int,
                  INSERTION: str, SEGMENT_ID: str}
        sort_keys = [ATOM_NUMBER]

        centrality_pdb_file_path = os.path.join(os.path.join(str(self.destination_output_path), self.pdb_name),
                                                centrality_csv_file)
        centrality_score_df = utils.get_df(output_file_path=centrality_pdb_file_path,
                                           columns=columns,
                                           dtypes=dtypes,
                                           sort_keys=sort_keys)

        centrality_scores = centrality_score_df[CENTRALITY_SCORE].to_numpy(dtype=np.float64)

        threshold = np.percentile(centrality_scores, 100 - quantile_percentage)

        logging.info(f'Calculated threshold value for {log_detail_name} : {threshold}')
        out_df = centrality_score_df[centrality_score_df[CENTRALITY_SCORE] >= threshold]
        out_path = os.path.join(self.destination_output_path, self.pdb_name, high_percentage_file)
        out_df.loc[:, INSERTION] = out_df[INSERTION].apply(lambda x: "''" if x == "" else x)
        out_df.loc[:, SEGMENT_ID] = out_df[SEGMENT_ID].apply(lambda x: "''" if x == "" else x)

        out_df.to_csv(path_or_buf=out_path,
                      sep=';',
                      index=False,
                      header=False,
                      float_format="%.7f",
                      columns=columns)

        self._create_pymol_data(out_df=out_df, centrality_type=centrality_type)

    @log_time("Centrality Analyzer - Create PyMol Data")
    def _create_pymol_data(self, out_df, centrality_type: CentralityType):
        """ Creates pymol query strings to be saved in a txt file for later use and also create pymol command to visualize
        residues with top quantile percentage centrality scores
        Parameters:
        -----------
        out_df: pd.DataFrame
            Data frame containing the data to be saved
        centrality_type: CentralityType
            can be the following: CentralityType.BET, CentralityType.CLOS, and CentralityType.DEG
        """
        self.pymol_utils.generate_pymol_query_string(out_df=out_df, centrality_type=centrality_type)

        original_pdb = f'{self.pdb_name}_original_{CENTRALITY_PDB_TEMPLATE.format(type=centrality_type.display_name)}'

        residue_list = list(
            zip(out_df[RESIDUE_NAME], out_df[CHAIN_ID], out_df[RESIDUE_NUMBER], out_df[INSERTION], out_df[SEGMENT_ID]))

        self.pymol_utils.export_pymol_script(
            full_path_to_pdb=os.path.join(self.destination_output_path, self.pdb_name, original_pdb),
            residues=residue_list,
            filename=f'original_{centrality_type.display_name}_{TOP_PERCENTAGE_TILE}',
            is_actual_residue=True)

        self.pymol_utils.export_pymol_script(
            full_path_to_pdb=os.path.join(self.destination_output_path, self.pdb_name, f'{self.pdb_name}.pdb'),
            residues=residue_list,
            filename=f'{centrality_type.display_name}_{TOP_PERCENTAGE_TILE}',
            is_actual_residue=True)

        self.pymol_utils.export_pymol_script(
            full_path_to_pdb=os.path.join(self.destination_output_path, self.pdb_name, f'{self.pdb_name}_pre.pdb'),
            residues=residue_list,
            filename=f'{centrality_type.display_name}_{TOP_PERCENTAGE_TILE}_pre',
            is_actual_residue=True)

    def _generate_quantile_setting_dict(self):
        """ Generate a quantile setting dictionary for quantile calculation.

        Returns
        -------
        Settings for each centrality type
        """
        return {
            CentralityType.BET: {
                PDB_FILE: self._get_pdb_file(centrality_type=CentralityType.BET),
                CSV_FILE: self._get_csv_file(centrality_type=CentralityType.BET),
                OUTFILE: self._get_outfile(centrality_type=CentralityType.BET),
                QUANTILE_PERCENTAGE: self.calculation_options[CentralityType.BET.display_name][VALUE]
            },
            CentralityType.CLOS: {
                PDB_FILE: self._get_pdb_file(centrality_type=CentralityType.CLOS),
                CSV_FILE: self._get_csv_file(centrality_type=CentralityType.CLOS),
                OUTFILE: self._get_outfile(centrality_type=CentralityType.CLOS),
                QUANTILE_PERCENTAGE: self.calculation_options[CentralityType.CLOS.display_name][VALUE]
            },
            CentralityType.DEG: {
                PDB_FILE: self._get_pdb_file(centrality_type=CentralityType.DEG),
                CSV_FILE: self._get_csv_file(centrality_type=CentralityType.DEG),
                OUTFILE: self._get_outfile(centrality_type=CentralityType.DEG),
                QUANTILE_PERCENTAGE: self.calculation_options[CentralityType.DEG.display_name][VALUE]
            }
        }

    def _get_pdb_file(self, centrality_type: CentralityType) -> str:
        """returns formatted pdb file name based on centrality type"""
        return f'{self.pdb_name}_{CENTRALITY_PDB_TEMPLATE.format(type=centrality_type.display_name)}'

    def _get_csv_file(self, centrality_type: CentralityType) -> str:
        """returns formatted pdb file name based on centrality type"""
        return f'{self.pdb_name}_{CENTRALITY_CSV_TEMPLATE.format(type=centrality_type.display_name)}'

    def _get_outfile(self, centrality_type: CentralityType) -> str:
        """returns formatted outfile name for high percentage based on centrality type"""
        return f'{self.pdb_name}_{HIGH_PERCENTAGE_TEMPLATE.format(type=centrality_type.display_name)}'

    def _get_cvs_filename(self, centrality_type: CentralityType) -> str:
        """returns formatted CSV filename based on centrality type"""
        return f'{self.pdb_name}_{CENTRALITY_CSV_TEMPLATE.format(type=centrality_type.display_name)}'

    def _calculate_quantile(self, centrality_type: CentralityType):
        """ Calculates the quantile value for the given centrality_type

        Parameters:
        -----------
        centrality_type : CentralityType
            centrality_type can be CentralityType.CLOS or CentralityType.DEG or CentralityType.BET
        """
        quantile_setting_dict = self._generate_quantile_setting_dict()
        self._compute_high_percentage_quantile(
            centrality_type=centrality_type,
            centrality_csv_file=quantile_setting_dict[centrality_type][CSV_FILE],
            high_percentage_file=quantile_setting_dict[centrality_type][OUTFILE],
            quantile_percentage=quantile_setting_dict[centrality_type][QUANTILE_PERCENTAGE])

    def _get_input_files(self) -> list:
        input_files = []
        if self._is_bet_checked():
            input_files.append(CentralityType.BET)

        if self._is_clos_checked():
            input_files.append(CentralityType.CLOS)

        if self._is_deg_checked():
            input_files.append(CentralityType.DEG)

        return input_files

    @staticmethod
    def _get_num_workers(input_files: list, log_msg: str) -> int:
        cpu_count = mp.cpu_count()
        num_workers = min(cpu_count, len(input_files))
        logging.info(f"{log_msg} | "
                     f"CPU cores detected: {cpu_count} | "
                     f"Input files: {len(input_files)} | "
                     f"Workers used: {num_workers}")
        return num_workers

    def calculate_quantiles(self, use_parallel: bool = False):
        """ Calculates the quantile values for the three centrality types

        Parameters:
        -----------
        use_parallel : bool, default False
            if it is true, it uses multiprocessing to calculate the quantiles for each centrality type.
            Otherwise, it calculates the quantiles for each centrality sequentially.
            The quantile results for each centrality type are saved in files.
        """
        if use_parallel:
            input_files = self._get_input_files()
            with mp.Pool(self._get_num_workers(input_files=input_files, log_msg="calculate quantiles")) as pool:
                pool.map(self._calculate_quantile, input_files)
        else:
            if self._is_bet_checked():
                self._compute_high_percentage_quantile(
                    centrality_type=CentralityType.BET,
                    centrality_csv_file=self._get_csv_file(centrality_type=CentralityType.BET),
                    high_percentage_file=self._get_outfile(centrality_type=CentralityType.BET),
                    quantile_percentage=self.calculation_options[CentralityType.BET.display_name][VALUE])

            if self._is_clos_checked():
                self._compute_high_percentage_quantile(
                    centrality_type=CentralityType.CLOS,
                    centrality_csv_file=self._get_csv_file(centrality_type=CentralityType.CLOS),
                    high_percentage_file=self._get_outfile(centrality_type=CentralityType.CLOS),
                    quantile_percentage=self.calculation_options[CentralityType.CLOS.display_name][VALUE])

            if self._is_deg_checked():
                self._compute_high_percentage_quantile(
                    centrality_type=CentralityType.DEG,
                    centrality_csv_file=self._get_csv_file(centrality_type=CentralityType.DEG),
                    high_percentage_file=self._get_outfile(centrality_type=CentralityType.DEG),
                    quantile_percentage=self.calculation_options[CentralityType.DEG.display_name][VALUE])

    def _is_deg_checked(self) -> bool:
        """ returns true if the centrality type degree is checked"""
        return self.calculation_options[CentralityType.DEG.display_name][IS_CHECKED]

    def _is_clos_checked(self) -> bool:
        """ returns true if the centrality type closeness is checked"""
        return self.calculation_options[CentralityType.CLOS.display_name][IS_CHECKED]

    def _is_bet_checked(self) -> bool:
        """ returns true if the centrality type betweenness is checked"""
        return self.calculation_options[CentralityType.BET.display_name][IS_CHECKED]

    def calculate_all_scores(self, use_parallel: bool = True):
        """ Calculates the centrality scores for betweenness, closeness, and degree of the generated graph based on
        the local interaction strength.

        Parameters:
        -----------
        use_parallel : bool, default False
            if it is true, it uses multiprocessing to calculate the centrality scores
            for betweenness, closeness, and degree. Otherwise, it calculates the centrality score for each sequentially.
        """
        if use_parallel:
            centrality_types = self._get_input_files()
            with mp.Pool(
                    self._get_num_workers(input_files=centrality_types, log_msg="calculate centrality scores")) as pool:
                pool.map(self._calculate_centrality_score, centrality_types)
        else:
            if self._is_bet_checked():
                self._calculate_centrality_score(centrality_type=CentralityType.BET)
            if self._is_clos_checked():
                self._calculate_centrality_score(centrality_type=CentralityType.CLOS)
            if self._is_deg_checked():
                self._calculate_centrality_score(centrality_type=CentralityType.DEG)

    @staticmethod
    def _centrality_message(*args, **kwargs):
        """ Generate the log message for _calculate_centrality_score method of the CentralityCalculation class based on
        given args and kwargs.
        Parameters:
        -----------
        *args : tuple
        **kwargs : dict

        Return:
        -----------
        The generated log message based on the centrality type.
        """
        centrality_type = kwargs.get(CENTRALITY_TYPE_KEY, None)
        if centrality_type is None and len(args) > 1:
            centrality_type = args[1]
        if centrality_type is None:
            centrality_type = CentralityType.BET
        if CentralityType.BET == centrality_type:
            centrality_type_name = CentralityType.BET.display_name_capitalized()
        elif CentralityType.CLOS == centrality_type:
            centrality_type_name = CentralityType.CLOS.display_name_capitalized()
        else:
            centrality_type_name = CentralityType.DEG.display_name_capitalized()

        return f"Centrality score calculation for {centrality_type_name}"

    @log_details(_centrality_message)
    @log_time("Centrality Analyzer - Calculate Centrality Scores")
    def _calculate_centrality_score(self, centrality_type: CentralityType = CentralityType.BET):
        r""" Calculate the centrality score for the given centrality type.

        Parameters:
        -----------
        centrality_type : CentralityType, default CentralityType.BET
        The centrality type to calculate the score from the generated graph.
        Also, necessary figures are generated from the centrality score.
        These figures are:
            2D plot of centrality score on the graph
            2D plot of centrality score for each residue index
            Histogram of centrality score for each centrality type
        In addition, the degree scores are plotted on the additional calculation like laplacian.
        Check out the details of plot_graph_2d_degree and plot_graph_2d_degree_with_coords in GraphPlotter class.
        """
        centrality_root_path = os.path.join(str(self.destination_output_path), str(self.pdb_name))
        utils.create_folder_not_exists(centrality_root_path)
        if CentralityType.BET == centrality_type:
            filename = self._get_cvs_filename(centrality_type=centrality_type)
            filename_path = os.path.join(centrality_root_path, filename)
            scores_dict = nx.betweenness_centrality(G=self.graph, weight="weight")
            utils.write_centrality_scores_to_file(
                filename_path=filename_path,
                score_dict=scores_dict,
                actual_residue_number_map=self.actual_residue_number_map)

            logging.info(f"{filename_path} has been saved to the disk!")

            self.graph_plotter.plot_graph_2d(graph=self.graph,
                                             filename="betweenness_graph",
                                             centrality_scores=dict(sorted(scores_dict.items())),
                                             title="Betweenness Centrality Scores",
                                             cmap='jet',
                                             node_size=60,
                                             alpha=1,
                                             font_size=6,
                                             font_color='black')

            self.graph_plotter.plot_figures(scores_dict=scores_dict,
                                            centrality_type_name=CentralityType.BET.display_name)

            self.graph_plotter.plot_histogram(scores_dict=scores_dict,
                                              centrality_type_name=CentralityType.BET.display_name)

        elif CentralityType.CLOS == centrality_type:
            filename = self._get_cvs_filename(centrality_type=centrality_type)
            filename_path = os.path.join(centrality_root_path, filename)
            scores_dict = nx.closeness_centrality(G=self.graph, distance='weight')
            utils.write_centrality_scores_to_file(
                filename_path=filename_path,
                score_dict=scores_dict,
                actual_residue_number_map=self.actual_residue_number_map)

            logging.info(f"{filename_path} has been saved to the disk!")

            self.graph_plotter.plot_graph_2d(graph=self.graph,
                                             filename="closeness_graph",
                                             centrality_scores=dict(sorted(scores_dict.items())),
                                             title="Closeness Centrality Scores",
                                             cmap='jet',
                                             node_size=60,
                                             alpha=1,
                                             font_size=6,
                                             font_color='black')

            self.graph_plotter.plot_figures(scores_dict=scores_dict,
                                            centrality_type_name=CentralityType.CLOS.display_name)

            self.graph_plotter.plot_histogram(scores_dict=scores_dict,
                                              centrality_type_name=CentralityType.CLOS.display_name)
        else:
            filename = self._get_cvs_filename(centrality_type=CentralityType.DEG)
            filename_path = os.path.join(centrality_root_path, filename)
            scores_dict = dict(self.graph.degree())
            utils.write_centrality_scores_to_file(
                filename_path=filename_path,
                score_dict=scores_dict,
                actual_residue_number_map=self.actual_residue_number_map,
                float_formatter='%8.5f')

            logging.info(f"{filename_path} has been saved to the disk!")

            self.graph_plotter.plot_graph_2d_degree(graph=self.graph, title="Degree")

            self.graph_plotter.plot_graph_2d_degree_with_coords(graph=self.graph,
                                                                title="Degree Centrality Scores")

            self.graph_plotter.plot_figures(scores_dict=dict(self.graph.degree()),
                                            centrality_type_name=CentralityType.DEG.display_name,
                                            use_decimal=False)

            self.graph_plotter.plot_histogram(scores_dict=dict(self.graph.degree()),
                                              centrality_type_name=CentralityType.DEG.display_name)

    def _get_structure_state(self, attr_name: str = RECORD_NAME) -> str:
        """ returns 'holo' if any node in the graph has record_name == 'HETATM', otherwise returns 'apo'. """
        for _, attrs in self.graph.nodes(data=True):
            if HETATM == attrs.get(attr_name):
                return HOLO
        return APO

    @log_time("Centrality Analyzer - Build Network")
    def build_network(self) -> nx.Graph:
        """ Builds a graph using networkx from contact and coordinate data.

        This method constructs a graph where each node represents a residue in a protein structure.
        Nodes are added with various attributes such as coordinates (x, y, z), chain id, residue number,
        and insertion information. Edges are added based on the provided contact data, with weights representing
        the interaction strength between residues.

        Returns
        -------
        None
            This method modifies the `self.graph` object in place, adding nodes and edges.

        Process:
        --------
        1. Loads contact data (source, target residues, and interaction value).
        2. Loads coordinates for each residue (x, y, z) and additional information such as chain ID, residue number,
           and insertion code.
        3. Adds nodes with their respective attributes to the graph.
        4. Adds weighted edges based on the contact data.

        Notes
        -----
        - If any residue is missing in `actual_residue_number_map`, it is skipped with a warning.
        - The contact data edges are weighted using the inverse of the interaction value from the contact dataframe.
        """
        node_attrs = [X, Y, Z, CHAIN_ID, RESIDUE_NUMBER, INSERTION, SEGMENT_ID, RESIDUE_NAME, RECORD_NAME]
        residue_edges_file_path = os.path.join(self.destination_output_path, self.pdb_name,
                                               f'{self.pdb_name}_{RESIDUE_EDGES_FILE}')
        residue_average_coords_file_path = os.path.join(self.destination_output_path, self.pdb_name,
                                                        f'{self.pdb_name}_{RESIDUE_AVERAGE_COORDINATES_FILE}')

        residue_edges_df = utils.get_df(output_file_path=residue_edges_file_path,
                                        columns=[SOURCE_RESIDUE, TARGET_RESIDUE, VALUE],
                                        dtypes={SOURCE_RESIDUE: int, TARGET_RESIDUE: int, VALUE: float},
                                        sort_keys=[SOURCE_RESIDUE, TARGET_RESIDUE])

        residue_average_coords_df = utils.get_df(residue_average_coords_file_path,
                                                 columns=[RESIDUE_INDEX, X_COORD, Y_COORD, Z_COORD],
                                                 dtypes={RESIDUE_INDEX: int,
                                                         X_COORD: float,
                                                         Y_COORD: float,
                                                         Z_COORD: float},
                                                 sort_keys=[RESIDUE_INDEX])

        sources = residue_edges_df[SOURCE_RESIDUE].to_numpy(dtype=np.int32)
        targets = residue_edges_df[TARGET_RESIDUE].to_numpy(dtype=np.int32)
        weights = 1. / residue_edges_df[VALUE].to_numpy(dtype=np.float64)

        self.graph = nx.Graph(name=GRAPH_NAME)

        missing_residues = 0

        for residue_index, x_coord, y_coord, z_coord in zip(residue_average_coords_df[RESIDUE_INDEX],
                                                            residue_average_coords_df[X_COORD],
                                                            residue_average_coords_df[Y_COORD],
                                                            residue_average_coords_df[Z_COORD]):

            if residue_index not in self.actual_residue_number_map:
                logging.warning(f"Warning: residue {residue_index} not found in actual_residue_number_map.")
                missing_residues += 1
                continue

            residue_name, chain_id, residue_number, insertion, segment_id, record_name = self.actual_residue_number_map[
                residue_index]
            attrs = {key: value for key, value in
                     zip(node_attrs,
                         [x_coord, y_coord, z_coord, chain_id, residue_number, insertion, segment_id, residue_name,
                          record_name])}
            if residue_index not in self.graph:
                self.graph.add_node(residue_index, **attrs)

        if missing_residues:
            logging.warning(f"Skipped {missing_residues} residues due to missing metadata.")

        valid_edges = [
            (u, v, w) for u, v, w in zip(sources, targets, weights)
            if u in self.graph and v in self.graph
        ]

        self.graph.add_weighted_edges_from(valid_edges)

        self.number_of_nodes = self.graph.number_of_nodes()

        self.graph_plotter.plot_graph_2d(graph=self.graph,
                                         filename="graph",
                                         title="Residue Interaction Network",
                                         equal_axis=True)

        self.graph_plotter.plot_graph_3d(graph=self.graph,
                                         filename="graph_3D",
                                         title="Residue Interaction Network 3D")

        nx.write_graphml(self.graph, os.path.join(self.destination_output_path, self.pdb_name,
                                                  f'{self.pdb_name}_{self._get_structure_state()}_network.graphml'))
        return self.graph


def main():
    pass


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
"""
    __author__ = 'Zehra Sarica'
    __email__ = ['sarica16@itu.edu.tr','zehraacar559@gmail.com']
"""

import csv
import logging
import os.path
from pathlib import Path
from typing import Optional

import networkx as nx

from rinpy.constants import CHAIN_ID, SEGMENT_ID, INSERTION, RESIDUE_NUMBER, RESIDUE_NAME

_WEIGHT = 'weight'
_ALLOSTERIC_COUPLING_CSV = "allosteric_coupling.csv"
_SEQUENTIAL_EFFICIENCY_CSV = "communication_path_efficiency.csv"
_EMPTY_SPACE = ''
_VALIDATION_VALUES = [None, '', "''", '""']

_RESIDUE_KEY_TYPE = tuple[str, int, str, str]
_RESIDUE_PAIR_TYPE = tuple[_RESIDUE_KEY_TYPE, _RESIDUE_KEY_TYPE]


class CommunicationPathEfficiency:
    """ computes the communication path efficiency for the given source and target networks.
    it calculates the sequential efficiency and allosteric coupling between the source and target residues.

    Parameters
    -----------
     source_network : nx.Graph
        Residue interaction network graph of the source protein - apo or wild type.
    target_network : nx.Graph
        Residue interaction network graph of the target protein - holo or mutant type
    output_path : Path
        Directory path where the results will be saved.
    residue_pairs : list of tuple, optional
        List of residue pairs for shortest path analysis.

        Each element represents a (source_residue, target_residue) pair.

        Each residue is uniquely identified by:
            (chain_id, residue_number, insertion_code, segment_id)

        Format:
            [
                ((chain_id, residue_number, insertion_code, segment_id),
                 (chain_id, residue_number, insertion_code, segment_id)),
                ...
            ]

        Example:
            [
                (('A', 12, '', ''), ('A', 144, '', '')),
                (('A', 13, '', ''), ('A', 145, '', ''))
            ]
    source_pdb_name : str
        Name of the source PDB file.
    target_pdb_name : str
        Name of the target PDB file.
    """

    def __init__(self, source_network: nx.Graph, target_network: nx.Graph, output_path: Path,
                 residue_pairs: list[_RESIDUE_PAIR_TYPE], source_pdb_name: str,
                 target_pdb_name: str) -> None:
        self.source_network = source_network
        self.target_network = target_network
        self.output_path = output_path

        self.residue_pairs = residue_pairs

        self.source_pdb_name = source_pdb_name
        self.target_pdb_name = target_pdb_name

        self.source_seq_effi_score = None
        self.source_end_to_end_effi_score = None
        self.source_path_rest_int_effi_score = None
        self.target_seq_effi_score = None
        self.target_end_to_end_effi_score = None
        self.target_path_rest_int_effi_score = None

    @staticmethod
    def _format_residue_pair_item(key: _RESIDUE_KEY_TYPE) -> str:
        chain_id, residue_number, insertion, segment_id = key
        insertion = (insertion or "").strip()
        segment_id = (segment_id or "").strip()
        name = f"{chain_id}{int(residue_number)}"
        if insertion:
            name += insertion
        if segment_id:
            name += f"({segment_id})"
        return name

    def _residue_pair_name(self, residue_pair: _RESIDUE_PAIR_TYPE) -> str:
        source, target = residue_pair
        return f"{self._format_residue_pair_item(source)}_{self._format_residue_pair_item(target)}"

    def compute(self) -> None:
        """ computes the sequential efficiency, end-to-end efficiency, path restricted internal efficiency scores
            of source and target networks. Also, computes the allosteric coupling scores of source and target networks
            for the given residue pairs.
        """
        for residue_pair in self.residue_pairs:
            source_results = self._compute_communication_path_efficiency(network=self.source_network,
                                                                         residue_pair=residue_pair)

            target_results = self._compute_communication_path_efficiency(network=self.target_network,
                                                                         residue_pair=residue_pair)
            if source_results is None or target_results is None:
                continue

            save_name = self._residue_pair_name(residue_pair=residue_pair)

            self.source_seq_effi_score, self.source_end_to_end_effi_score, self.source_path_rest_int_effi_score = source_results
            self.target_seq_effi_score, self.target_end_to_end_effi_score, self.target_path_rest_int_effi_score = target_results

            sequential_efficiency_score_map = {
                f'{self.source_pdb_name}_sequential_efficiency_score': self.source_seq_effi_score,
                f'{self.source_pdb_name}_end_to_end_efficiency_score': self.source_end_to_end_effi_score,
                f'{self.source_pdb_name}_path_restricted_efficiency_score': self.source_path_rest_int_effi_score,
                f'{self.target_pdb_name}_sequential_efficiency_score': self.target_seq_effi_score,
                f'{self.target_pdb_name}_end_to_end_efficiency_score': self.target_end_to_end_effi_score,
                f'{self.target_pdb_name}_path_restricted_efficiency_score': self.target_path_rest_int_effi_score
            }
            self._write_sequential_efficiency_scores_to_csv(sequential_efficiency_score_map, save_name)

            results = self._compute_allosteric_coupling(self.source_network, self.target_network, residue_pair)
            self._write_allosteric_coupling_scores_to_csv(results, save_name)

    def _write_sequential_efficiency_scores_to_csv(self, sequential_efficiency_score_map: dict[str, float],
                                                   save_name: str) -> None:
        """ save the sequential efficiency scores to a CSV file for the given results."""
        save_path = os.path.join(self.output_path, f"{save_name}_{_SEQUENTIAL_EFFICIENCY_CSV}")
        with open(save_path, "w", newline="") as f:
            writer = csv.writer(f)
            writer.writerow(["metric", "value"])
            for metric, value in sequential_efficiency_score_map.items():
                writer.writerow([metric, f"{value:.6f}"])
        logging.info(f"sequential efficiency scores has been saved to {save_path}.")

    def _write_allosteric_coupling_scores_to_csv(self, results: dict[str, dict[str, float]], save_name: str) -> None:
        """ save the allosteric coupling scores to a CSV file for the given results."""
        save_path = os.path.join(self.output_path, f"{save_name}_{_ALLOSTERIC_COUPLING_CSV}")
        with open(save_path, "w", newline="") as f:
            writer = csv.writer(f)
            writer.writerow(["category", "metric", "value"])
            for category, metrics in results.items():
                for metric, value in metrics.items():
                    writer.writerow([category, metric, f"{value:.6f}"])

        logging.info(f"coupling scores has been saved to {save_path}.")

    @staticmethod
    def build_node_map(network: nx.Graph) -> dict[tuple[str, int, str, str], int]:
        """build node map for the given network."""
        node_map = {}
        for node, data in network.nodes(data=True):
            chain_id = str(data.get(CHAIN_ID))
            residue_number = int(data.get(RESIDUE_NUMBER))
            insertion = data.get(INSERTION, _EMPTY_SPACE)
            if str(insertion) in _VALIDATION_VALUES:
                insertion = _EMPTY_SPACE
            segment_id = data.get(SEGMENT_ID, _EMPTY_SPACE)
            if str(segment_id) in _VALIDATION_VALUES:
                segment_id = _EMPTY_SPACE
            key = (chain_id, residue_number, str(insertion), str(segment_id))
            node_map[key] = node
        return node_map

    @staticmethod
    def find_node(node_map: dict, residue_tuple: _RESIDUE_KEY_TYPE) -> dict:
        """ find the node with the given residue key tuple."""
        chain_id, residue_number, insertion, segment_id = residue_tuple
        if str(insertion) in _VALIDATION_VALUES:
            insertion = _EMPTY_SPACE
        if str(segment_id) in _VALIDATION_VALUES:
            segment_id = _EMPTY_SPACE

        key = (str(chain_id), int(residue_number), str(insertion), str(segment_id))
        return node_map.get(key)

    def _save_shortest_path_residues(self, network: nx.Graph, shortest_path_nodes: list[int],
                                     residue_pair: _RESIDUE_PAIR_TYPE):
        """ save the shortest path residues into a txt file"""
        if not shortest_path_nodes:
            return
        save_name = self._residue_pair_name(residue_pair=residue_pair)
        output_file = os.path.join(self.output_path, f"{save_name}_shortest_path_residues.txt")
        with open(output_file, "w") as f:
            for node_id in shortest_path_nodes:
                residue_name = network.nodes[node_id].get(RESIDUE_NAME, "")
                chain_id = network.nodes[node_id].get(CHAIN_ID, "")
                residue_number = network.nodes[node_id].get(RESIDUE_NUMBER, "")
                insertion = network.nodes[node_id].get(INSERTION, "")
                segment_id = network.nodes[node_id].get(SEGMENT_ID, "")
                insertion = insertion if insertion else "''"
                segment_id = segment_id if segment_id else "''"
                f.write(f"{residue_name}, {chain_id}, {residue_number}, {insertion}, {segment_id}\n")

    def _compute_communication_path_efficiency(self, network: nx.Graph, residue_pair: _RESIDUE_PAIR_TYPE) -> Optional[
        tuple[float, float, float]]:
        """ compute sequential efficiency of the network for the given residue pairs. """
        node_map = self.build_node_map(network)

        sequential_efficiency_values = []
        end_to_end_efficiency_values = []
        path_restricted_efficiency_values = []

        source_residue, target_residue = residue_pair

        source_node = self.find_node(node_map=node_map, residue_tuple=source_residue)
        target_node = self.find_node(node_map=node_map, residue_tuple=target_residue)
        if source_node is None or target_node is None:
            logging.warning(f"Nodes not found: {source_residue}-{target_residue}")
            return None
        try:
            path = nx.shortest_path(network, source_node, target_node, weight=_WEIGHT)
            self._save_shortest_path_residues(network, path, residue_pair)
            k = len(path)
            edge_count = k - 1

            if edge_count <= 0:
                return None

            weight_sum = 0
            for i in range(edge_count):
                u = path[i]
                v = path[i + 1]
                weight_sum += network[u][v].get(_WEIGHT, 1.0)

            if weight_sum == 0:
                return None

            # sequential efficiency
            efficiency_seq = edge_count / weight_sum
            sequential_efficiency_values.append(efficiency_seq)

            # End-to-end efficiency
            efficiency_end = 1 / weight_sum
            end_to_end_efficiency_values.append(efficiency_end)

            #  path restricted internal efficiency
            inverse_dist_sum = 0
            pair_count = 0

            for i in range(k):
                for j in range(i + 1, k):
                    try:
                        d = nx.shortest_path_length(network, path[i], path[j], weight=_WEIGHT)
                        if d and d > 0:
                            inverse_dist_sum += 1 / d
                            pair_count += 1
                    except nx.NetworkXNoPath:
                        continue
            if k > 1:
                path_efficiency = (2 / (k * (k - 1))) * inverse_dist_sum
                path_restricted_efficiency_values.append(path_efficiency)

        except nx.NetworkXNoPath:
            logging.error(f"No path found: {source_residue}-{target_residue}")
            return None

        seq_avg = sum(sequential_efficiency_values) / len(
            sequential_efficiency_values) if sequential_efficiency_values else 0
        end_avg = sum(end_to_end_efficiency_values) / len(
            end_to_end_efficiency_values) if end_to_end_efficiency_values else 0
        path_avg = sum(path_restricted_efficiency_values) / len(
            path_restricted_efficiency_values) if path_restricted_efficiency_values else 0

        return seq_avg, end_avg, path_avg

    def _compute_allosteric_coupling(self, source_network: nx.Graph, target_network: nx.Graph,
                                     residue_pair: _RESIDUE_PAIR_TYPE) -> dict[str, dict[str, float]]:
        """computes allosteric coupling metrics of the source and target network for the given pairs."""
        source_seq, source_end, source_path = self._compute_communication_path_efficiency(source_network, residue_pair)
        target_seq, target_end, target_path = self._compute_communication_path_efficiency(target_network, residue_pair)

        coupling_end_to_end = target_end - source_end
        coupling_path_restricted = target_path - source_path

        results = {
            f"{self.source_pdb_name}": {
                "end-to-end_efficiency": source_end,
                "path_restricted_internal_efficiency": source_path
            },
            f"{self.target_pdb_name}": {
                "end-to-end_efficiency": target_end,
                "path_restricted_internal_efficiency": target_path
            },
            "allosteric_coupling": {
                "end_to_end_efficiency": coupling_end_to_end,
                "path_restricted_internal_efficiency": coupling_path_restricted
            }
        }
        return results


def main():
    pass


if __name__ == '__main__':
    main()
